############################################################################
# tools/pynuttx/nxgdb/fs/inode.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import enum
from typing import Generator, Tuple, Union

import gdb
import nxgdb.protocols.fs as p
import nxgdb.utils as utils
from nxgdb.fs.utils import (
    CONFIG_PSEUDOFS_ATTRIBUTES,
    CONFIG_PSEUDOFS_FILE,
    FSNODEFLAG_TYPE_MASK,
)

g_special_inodes = {}  # Map of the special inodes including epoll, inotify, etc.


class InodeType(enum.Enum):
    # define   FSNODEFLAG_TYPE_PSEUDODIR  0x00000000 /*   Pseudo dir (default)   */
    # define   FSNODEFLAG_TYPE_DRIVER     0x00000001 /*   Character driver       */
    # define   FSNODEFLAG_TYPE_BLOCK      0x00000002 /*   Block driver           */
    # define   FSNODEFLAG_TYPE_MOUNTPT    0x00000003 /*   Mount point            */
    # define   FSNODEFLAG_TYPE_NAMEDSEM   0x00000004 /*   Named semaphore        */
    # define   FSNODEFLAG_TYPE_MQUEUE     0x00000005 /*   Message Queue          */
    # define   FSNODEFLAG_TYPE_SHM        0x00000006 /*   Shared memory region   */
    # define   FSNODEFLAG_TYPE_MTD        0x00000007 /*   Named MTD driver       */
    # define   FSNODEFLAG_TYPE_SOFTLINK   0x00000008 /*   Soft link              */
    # define   FSNODEFLAG_TYPE_SOCKET     0x00000009 /*   Socket                 */
    # define   FSNODEFLAG_TYPE_PIPE       0x0000000a /*   Pipe                   */
    # define   FSNODEFLAG_TYPE_NAMEDEVENT 0x0000000b /*   Named event group      */
    PSEUDODIR = 0
    DRIVER = 1
    BLOCK = 2
    MOUNTPT = 3
    NAMEDSEM = 4
    MQUEUE = 5
    SHM = 6
    MTD = 7
    SOFTLINK = 8
    SOCKET = 9
    PIPE = 10
    NAMEDEVENT = 11
    UNKNOWN = 12


class Inode(utils.Value, p.Inode):
    def __init__(self, obj: Union[gdb.Value, utils.Value]):
        super().__init__(obj)

    def __str__(self) -> str:
        name = get_inode_name(self)
        type = inode_gettype(self)
        return f"{name}{'/' if type == InodeType.PSEUDODIR else ''}{int(self): #x} {type.name}"

    def __repr__(self) -> str:
        return self.__str__()

    def details(self) -> str:
        details = (
            f"ino: {self.i_ino}, crefs: {self.i_crefs}"
            f", flags: {self.i_flags}, private: {self.i_private}"
        )

        if CONFIG_PSEUDOFS_FILE:
            details += f", i_size: {self.i_size}"
        if CONFIG_PSEUDOFS_ATTRIBUTES:
            details += f", i_mode: {self.i_mode}, i_owner: {self.i_owner}, i_group: {self.i_group}"
            details += f", i_atime: {self.i_atime}, i_mtime: {self.i_mtime}, i_ctime: {self.i_ctime}"
        return details

    @property
    def nodetype(self) -> InodeType:
        return inode_gettype(self)


def inode_gettype(inode: p.Inode) -> InodeType:
    if not inode:
        return InodeType.UNKNOWN

    type = int(inode.i_flags & FSNODEFLAG_TYPE_MASK)

    # check if it's a valid type in InodeType
    if type in [e.value for e in InodeType]:
        return InodeType(type)

    return InodeType.UNKNOWN


def get_inode_name(inode: p.Inode):
    try:
        if not inode:
            return ""

        def special_inode_name(inode):
            global g_special_inodes
            if not g_special_inodes:
                g_special_inodes = {}
                for special in (
                    "perf",
                    "timerfd",
                    "signalfd",
                    "dir",
                    "inotify",
                    "epoll",
                    "eventfd",
                    "sock",
                ):
                    value = utils.gdb_eval_or_none(f"g_{special}_inode")
                    if value:
                        g_special_inodes[special] = value.address

            for name, value in g_special_inodes.items():
                if value == inode:
                    return name

            return None

        if name := special_inode_name(inode):
            return name

        ptr = inode.i_name.cast(gdb.lookup_type("char").pointer())
        return ptr.string()
    except gdb.MemoryError:
        return "<MemoryError>"


def inode_getpath(inode: p.Inode):
    """get path from inode"""
    try:
        if not inode:
            return ""

        name = get_inode_name(inode)

        if inode.i_parent:
            return inode_getpath(inode.i_parent) + "/" + name

        return name
    except gdb.MemoryError:
        return "<MemoryError>"


def get_root_inode(addr_or_expr):
    try:
        return utils.Value(int(addr_or_expr, 0)).cast(
            gdb.lookup_type("struct inode").pointer()
        )
    except ValueError:
        return utils.gdb_eval_or_none(addr_or_expr)


def print_inode_info(
    nodes, node: Inode, level=1, prefix="", maxlevel=4096, type=None, verbose=False
):
    if level > maxlevel:
        return

    while node:
        if node.i_peer:
            initial_indent = prefix + "├── "
            newprefix = prefix + "│   "
        else:
            initial_indent = prefix + "└── "
            newprefix = prefix + "    "

        if (
            not type
            or node.nodetype == type
            or any(n.i_parent == node for n, _ in nodes)
        ):
            gdb.write(f"{initial_indent}{node} {node.details() if verbose else ''}\n")

        if node.i_child:
            print_inode_info(
                nodes,
                Inode(node.i_child),
                level + 1,
                newprefix,
                maxlevel,
                type,
                verbose,
            )
        node = Inode(node.i_peer)


def foreach_inode(root=None, path="") -> Generator[Tuple[p.Inode, str], None, None]:
    node: p.Inode = root or utils.parse_and_eval("g_root_inode").i_child
    while node:
        newpath = path + "/" + get_inode_name(node)
        yield node, newpath
        if node.i_child:
            yield from foreach_inode(node.i_child, newpath)
        node = node.i_peer


def get_fstype(node: p.Inode):
    try:
        if not node:
            return ""

        statfs = node.u.i_mops.statfs
        funcname = gdb.block_for_pc(int(statfs)).function.print_name
        return funcname.split("_")[0]

    except gdb.MemoryError:
        return "<MemoryError>"


def fstype_filter(fstype):
    """Filter inodes by filesystem type"""

    def filter(item):
        inode, path = item
        if inode_gettype(inode) != InodeType.MOUNTPT:
            return False
        return get_fstype(inode) == fstype

    return filter



############################################################################
# tools/pynuttx/nxgdb/memclassifier/size.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

categories = {
    "l0-100": lambda mb: mb.block_size() <= 100,
    "l100-500": lambda mb: mb.block_size() <= 500,
    "l500_1000": lambda mb: mb.block_size() <= 1000,
    "l1000_2000": lambda mb: mb.block_size() <= 2000,
    "l2000_4000": lambda mb: mb.block_size() <= 4000,
    "l4000+": lambda mb: True,
}



############################################################################
# tools/pynuttx/nxgdb/lists.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse

import gdb

from . import autocompeletion, utils

list_node_type = utils.lookup_type("struct list_node")
sq_queue_type = utils.lookup_type("sq_queue_t")
sq_entry_type = utils.lookup_type("sq_entry_t")
dq_queue_type = utils.lookup_type("dq_queue_t")


class NxList:
    def __init__(self, list, container_type=None, member=None, reverse=False):
        """Initialize the list iterator. Optionally specify the container type and member name."""

        if not list:
            raise ValueError("The head cannot be None.\n")

        if list.type.code != gdb.TYPE_CODE_PTR:
            list = list.address  # Make sure list is a pointer.

        if container_type and not member:
            raise ValueError("Must specify the member name in container.\n")

        self.list = list
        self.reverse = reverse
        self.container_type = container_type
        self.member = member
        self.current = self._get_first()

    def _get_first(self):
        """Get the initial node based on the direction of traversal."""

        prev = self.list["prev"]
        next = self.list["next"]

        first = prev if self.reverse else next
        return first if first and first != self.list else None

    def _get_next(self, node):
        #   for(node = (list)->next; node != (list); node = node->next)
        return node["next"] if node["next"] != self.list else None

    def _get_prev(self, node):
        #   for(node = (list)->next; node != (list); node = node->prev)
        return node["prev"] if node["prev"] != self.list else None

    def __iter__(self):
        return self

    def __next__(self):
        if self.current is None:
            raise StopIteration

        node = self.current
        self.current = self._get_prev(node) if self.reverse else self._get_next(node)
        return (
            utils.container_of(node, self.container_type, self.member)
            if self.container_type
            else node
        )


class NxSQueue(NxList):
    def __init__(self, list, container_type=None, member=None, reverse=False):
        """Initialize the singly linked list iterator. Optionally specify the container type and member name."""
        if reverse:
            raise ValueError(
                "Reverse iteration is not supported for singly linked lists.\n"
            )
        super().__init__(list, container_type, member, reverse)

    def _get_first(self):
        #   for ((p) = (q)->head; (p) != NULL; (p) = (p)->flink)
        return self.list["head"] or None

    def _get_next(self, node):
        # if not node["flink"], then return None, to indicate end of list
        return node["flink"] or None


class NxDQueue(NxList):
    def __init__(self, list, container_type=None, member=None, reverse=False):
        """Initialize the doubly linked list iterator. Optionally specify the container type and member name."""
        super().__init__(list, container_type, member, reverse)

    def _get_first(self):
        head = self.list["head"]
        tail = self.list["tail"]

        first = head if not self.reverse else tail
        return first or None

    def _get_next(self, node):
        #   for ((p) = (q)->head; (p) != NULL; (p) = (p)->flink)
        return node["flink"] or None

    def _get_prev(self, node):
        #   for ((p) = (q)->tail; (p) != NULL; (p) = (p)->blink)
        return node["blink"] or None


def list_check(head):
    """Check the consistency of a list"""
    nb = 0

    if head.type == list_node_type.pointer():
        head = head.dereference()
    elif head.type != list_node_type:
        raise gdb.GdbError("argument must be of type (struct list_node [*])")
    c = head
    try:
        gdb.write("Starting with: {}\n".format(c))
    except gdb.MemoryError:
        gdb.write("head is not accessible\n")
        return
    while True:
        p = c["prev"].dereference()
        n = c["next"].dereference()
        try:
            if p["next"] != c.address:
                gdb.write(
                    "prev.next != current: "
                    "current@{current_addr}={current} "
                    "prev@{p_addr}={p}\n".format(
                        current_addr=c.address,
                        current=c,
                        p_addr=p.address,
                        p=p,
                    )
                )
                return
        except gdb.MemoryError:
            gdb.write(
                "prev is not accessible: "
                "current@{current_addr}={current}\n".format(
                    current_addr=c.address, current=c
                )
            )
            return
        try:
            if n["prev"] != c.address:
                gdb.write(
                    "next.prev != current: "
                    "current@{current_addr}={current} "
                    "next@{n_addr}={n}\n".format(
                        current_addr=c.address,
                        current=c,
                        n_addr=n.address,
                        n=n,
                    )
                )
                return
        except gdb.MemoryError:
            gdb.write(
                "next is not accessible: "
                "current@{current_addr}={current}\n".format(
                    current_addr=c.address, current=c
                )
            )
            return
        c = n
        nb += 1
        if c == head:
            gdb.write("list is consistent: {} node(s)\n".format(nb))
            return


def sq_is_empty(sq):
    """Check if a singly linked list is empty"""
    if sq.type == sq_queue_type.pointer():
        sq = sq.dereference()
    elif sq.type != sq_queue_type:
        return False

    if sq["head"] == 0:
        return True
    else:
        return False


def sq_check(sq, verbose=True) -> int:
    """Check the consistency of a singly linked list"""
    nb = 0
    if sq.type == sq_queue_type.pointer():
        sq = sq.dereference()
    elif sq.type != sq_queue_type:
        gdb.write("Must be struct sq_queue not {}".format(sq.type))
        return nb

    if sq["head"] == 0:
        if verbose:
            gdb.write("sq_queue head is empty {}\n".format(sq.address))
        return nb

    nodes = set()
    entry = sq["head"].dereference()
    try:
        while entry.address:
            nb += 1
            if int(entry.address) in nodes:
                gdb.write("sq_queue is circular: {}\n".format(entry.address))
                return nb
            nodes.add(int(entry.address))
            entry = entry["flink"].dereference()
    except gdb.MemoryError:
        gdb.write("entry address is unaccessible {}\n".format(entry.address))
        return nb

    if int(sq["tail"]) not in nodes:
        gdb.write("sq_queue tail is not in the list {}\n".format(sq["tail"]))
        return nb
    if sq["tail"]["flink"] != 0:
        gdb.write("sq_queue tail->flink is not null {}\n".format(sq["tail"]))
        return nb

    if verbose:
        gdb.write("sq_queue is consistent: {} node(s)\n".format(nb))
    return nb


def sq_count(sq) -> int:
    """Count sq elements, abort if check failed"""
    return sq_check(sq, verbose=False)


def sq_for_every(sq, entry=None):
    """Iterate over a singly linked list"""
    if sq.type == sq_queue_type.pointer():
        sq = sq.dereference()
    elif sq.type != sq_queue_type:
        gdb.write("Must be struct sq_queue not {}".format(sq.type))
        return

    if sq["head"] == 0:
        return

    if not entry:
        entry = sq["head"].dereference()

    while entry.address:
        yield entry
        entry = entry["flink"].dereference()


def dq_for_every(dq, entry=None):
    """Iterate over a doubly linked list"""
    if dq.type == dq_queue_type.pointer():
        dq = dq.dereference()
    elif dq.type != dq_queue_type:
        gdb.write("Must be struct dq_queue not {}".format(dq.type))
        return

    if dq["head"] == 0:
        return

    if not entry:
        entry = dq["head"].dereference()

    while entry.address:
        yield entry.address
        entry = entry["flink"].dereference()


def dq_check(dq):
    """Check the consistency of a doubly linked list"""
    nb = 0
    if dq.type == dq_queue_type.pointer():
        dq = dq.dereference()
    elif dq.type != dq_queue_type:
        gdb.write("Must be struct dq_queue not {}".format(dq.type))
        return

    if dq["head"] == 0:
        gdb.write("dq_queue head is empty {}\n".format(dq.address))
        return

    nodes = set()
    entry = dq["head"].dereference()
    try:
        while entry.address:
            nb += 1
            if int(entry.address) in nodes:
                gdb.write("dq_queue is circular: {}\n".format(entry.address))
                return
            nodes.add(int(entry.address))
            entry = entry["flink"].dereference()
    except gdb.MemoryError:
        gdb.write("entry address is unaccessible {}\n".format(entry.address))
        return

    if int(dq["tail"]) not in nodes:
        gdb.write("dq_queue tail is not in the list {}\n".format(dq["tail"]))
        return
    if dq["tail"]["flink"] != 0:
        gdb.write("dq_queue tail->flink is not null {}\n".format(dq["tail"]))
        return

    gdb.write("dq_queue is consistent: {} node(s)\n".format(nb))


class ListCheck(gdb.Command):
    """Verify a list consistency"""

    def __init__(self):
        super().__init__("list_check", gdb.COMMAND_DATA, gdb.COMPLETE_EXPRESSION)

    @utils.dont_repeat_decorator
    def invoke(self, arg, from_tty):
        argv = gdb.string_to_argv(arg)
        if len(argv) != 1:
            raise gdb.GdbError("nx-list-check takes one argument")

        obj = utils.parse_and_eval(argv[0])
        if obj.type == list_node_type.pointer():
            list_check(obj)
        elif obj.type == sq_queue_type.pointer():
            sq_check(obj)
        elif obj.type == dq_queue_type.pointer():
            dq_check(obj)
        else:
            raise gdb.GdbError("Invalid argument type: {}".format(obj.type))


@autocompeletion.complete
class ForeachListEntry(gdb.Command):
    """Dump list members for a given list"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description="Iterate the items in list")
        parser.add_argument("head", type=str, help="List head", metavar="symbol")
        parser.add_argument(
            "-n",
            "--next",
            type=str,
            help="The name of the next pointer in the list node",
            default="next",
        )
        parser.add_argument(
            "-c", "--container", type=str, default=None, help="Optional container type"
        )
        parser.add_argument(
            "-m", "--member", type=str, default=None, help="Member name in container"
        )
        parser.add_argument(
            "-e",
            "--element",
            type=str,
            help="Only dump this element in array member struct.",
            default=None,
        )
        return parser

    def __init__(self):
        super().__init__("foreach list", gdb.COMMAND_DATA)
        utils.alias("list-foreach", "foreach list")
        self.parser = self.get_argparser()

    @utils.dont_repeat_decorator
    def invoke(self, arg, from_tty):
        argv = gdb.string_to_argv(arg)
        try:
            args = self.parser.parse_args(argv)
        except SystemExit:
            gdb.write("Invalid arguments\n")
            return

        pointer = utils.parse_and_eval(args.head)
        node = pointer
        i = 0
        while node:
            entry = (
                utils.container_of(node, args.container, args.member)
                if args.container
                else node
            )
            entry = entry.dereference()
            entry = entry[args.element] if args.element else entry
            gdb.write(
                f"{i} *({entry.type} *){hex(entry.address)} {entry.format_string(styling=True)}\n"
            )
            i += 1
            node = node[args.next]
            if node == pointer:
                break


@autocompeletion.complete
class ForeachArray(gdb.Command):
    """Dump array members."""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description="Iterate the items in array")
        parser.add_argument("head", type=str, help="List head")
        parser.add_argument(
            "-l",
            "--length",
            type=int,
            help="The array length",
            default=None,
        )
        parser.add_argument(
            "-e",
            "--element",
            type=str,
            help="Only dump this element in array member struct.",
            default=None,
        )
        return parser

    def __init__(self):
        super().__init__("foreach array", gdb.COMMAND_DATA)
        utils.alias("array-foreach", "foreach array")
        self.parser = self.get_argparser()

    @utils.dont_repeat_decorator
    def invoke(self, arg, from_tty):
        argv = gdb.string_to_argv(arg)
        try:
            args = self.parser.parse_args(argv)
        except SystemExit:
            gdb.write("Invalid arguments\n")
            return

        pointer = utils.parse_and_eval(args.head)
        node = pointer
        len = args.length if args.length else utils.nitems(pointer)
        for i in range(len):
            entry = node[i][args.element] if args.element else node[i]
            gdb.write(f"{i}: {entry.format_string(styling=True)}\n")



############################################################################
# tools/pynuttx/nxelf/macros.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

# NOTE: GDB stores macro information based on the current stack frame's scope,
# including the source file and line number. Therefore, there may be missing
# macro definitions when you are at different stack frames.
#
# To resolve this issue, we need to retrieve all macro information from the ELF file
# then parse and evaluate it by ourselves.
#
# There might be two ways to achieve this, one is to leverage the C preprocessor
# to directly preprocess all the macros interpreted into python constants
# gcc -E -x c -P <file_with_macros> -I/path/to/nuttx/include
#
# While the other way is to leverage the dwarf info stored in the ELF file,
# with -g3 switch, we have a `.debug_macro` section containing all the information
# about the macros.
#
# Currently, we are using the second method.

import hashlib
import json
import os
import re
import tempfile
import time
from os import path

PUNCTUATORS = [
    r"\[",
    r"\]",
    r"\(",
    r"\)",
    r"\{",
    r"\}",
    r"\?",
    r";",
    r",",
    r"~",
    r"\.\.\.",
    r"\.",
    r"\-\>",
    r"\-\-",
    r"\-\=",
    r"\-",
    r"\+\+",
    r"\+\=",
    r"\+",
    r"\*\=",
    r"\*",
    r"\!\=",
    r"\!",
    r"\&\&",
    r"\&\=",
    r"\&",
    r"\/\=",
    r"\/",
    r"\%\>",
    r"%:%:",
    r"%:",
    r"%=",
    r"%",
    r"\^\=",
    r"\^",
    r"\#\#",
    r"\#",
    r"\:\>",
    r"\:",
    r"\|\|",
    r"\|\=",
    r"\|",
    r"<<=",
    r"<<",
    r"<=",
    r"<:",
    r"<%",
    r"<",
    r">>=",
    r">>",
    r">=",
    r">",
    r"\=\=",
    r"\=",
]


def parse_macro(line, macros, pattern):
    # grep name, value
    # the first group matches the token, the second matches the replacement
    m = pattern.match(line)
    if not m:
        return False

    name, value = m.group(1), m.group(2)

    if name in macros:
        # FIXME: what should we do if we got a redefinition/duplication here?
        # for now I think it's ok just overwrite the old value
        pass

    # emplace, for all undefined macros we evalute it to zero
    macros[name] = value if value else "0"

    return True


def fetch_macro_info(file):
    if not path.isfile(file):
        raise FileNotFoundError("No given ELF target found")

    # FIXME: we don't use subprocess here because
    # it's broken on some GDB distribution :(, I haven't
    # found a solution to it.

    with open(file, "rb") as f:
        hash = hashlib.md5(f.read()).hexdigest()

    macros = {}
    p = re.compile(r".*macro[ ]*:[ ]*([\S]+\(.*?\)|[\w]+)[ ]*(.*)")
    cache = path.join(tempfile.gettempdir(), f"{hash}.json")
    print(f"Load macro: {cache}")
    if not path.isfile(cache):
        t = time.time()
        os.system(f'readelf -wm "{file}" > "{cache}"')
        print(f"readelf took {time.time() - t:.1f} seconds")

        t = time.time()
        with open(cache, "r") as f2:
            for line in f2.readlines():
                if not line.startswith(" DW_MACRO_define") and not line.startswith(
                    " DW_MACRO_undef"
                ):
                    continue

                if not parse_macro(line, macros, p):
                    print(f"Failed to parse {line}")

        print(f"Parse macro took {time.time() - t:.1f} seconds")

        with open(cache, "w") as f2:
            dump = json.dumps(macros, indent=4, sort_keys=True)
            f2.write(dump)

        print(f"Cache macro info to {cache}")
    else:
        with open(cache, "r") as f2:
            try:
                macros = json.load(f2)
            except json.decoder.JSONDecodeError:
                print(f"Warning: Failed to load {cache}, reprocessing...")
                os.remove(cache)
                return fetch_macro_info(file)

    return macros


def split_tokens(expr):
    p = "(" + "|".join(PUNCTUATORS) + ")"
    res = list(
        filter(lambda e: e != "", map(lambda e: e.rstrip().lstrip(), re.split(p, expr)))
    )
    return res


def do_expand(expr, macro_map):
    if expr in PUNCTUATORS:
        return expr

    tokens = split_tokens(expr)

    res = []

    for t in tokens:
        if t not in macro_map:
            res.append(t)
            continue
        res += do_expand(macro_map[t], macro_map)

    return res


# NOTE: Implement a fully functional parser which can
# preprocess all the C marcos according to ISO 9899 standard
# may be an overkill, what we really care about are those
# macros that can be evaluated to an constant value.
#
# #define A (B + C + D)
# #define B 1
# #define C 2
# #define D 3
# invoking try_expand('A', macro_map) will give you "(1 + 2 + 3)"
#
# However,
# #define SUM(B,C,D) (B + C + D)
# invoking try_expand('SUM(1,2,3)', macro_map) will give you "SUM(1,2,3)"
#
# We have not implemented this feature as we have not found a practical
# use case for it in our GDB plugin.
#
# However, you can switch to the correct stack frame that has this macro defined
# and let GDB expand and evaluate it for you if you really want to evaluate some very
# complex macros.


def try_expand(expr, macro):
    res = []

    res += do_expand(expr, macro)

    return "".join(res)


class Macro:
    """
    This is a singleton class which only initializes once to
    cache a context of macro definition which can be queried later
    TODO: we only deal with single ELF at the moment for simplicity
    If you load more object files while debugging, only the first one gets loaded
    will be used to retrieve macro information

    Usage:
        macro = Macro("nuttx/nuttx")
        print(macro.CONFIG_MM_RECORD_STACK)
        if macro.CONFIG_MM_RECORD_STACK:
            print("mm backtrace is enabled")
        else:
            print("mm backtrace is disabled")
    """

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "instance"):
            cls.instance = super(Macro, cls).__new__(cls)
        return cls.instance

    def __init__(self, filename):
        self._macro_map = {}
        self._file = filename
        self._macro_map = fetch_macro_info(filename)

    def is_defined(self, macro_name):
        """
        Check if a macro is defined
        """
        return macro_name in self._macro_map

    def get_value(self, macro_name, default=None):
        """
        Get the value of a macro, return default if macro is not defined
        """
        if not self.is_defined(macro_name):
            return default

        value = self._macro_map[macro_name]
        # Try to convert to numeric type
        try:
            # Handle hexadecimal
            if isinstance(value, str) and value.startswith("0x"):
                return int(value, 16)
            # Handle integer
            return int(value)
        except (ValueError, TypeError):
            # Return original value if conversion fails
            return value

    def __getattr__(self, name):
        """
        Allow using dot notation to access macros
        """
        return self.get_value(name)

    def __call__(self, macro_name):
        """
        Allow using function call syntax to get macro
        """
        return self.get_value(macro_name)



############################################################################
# tools/pynuttx/nxgdb/protocols/wqueue.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from __future__ import annotations

from typing import List

from .value import Value


class Work(Value):
    """struct work_s"""

    worker: Value  # void (*worker_t)(FAR void *arg);
    arg: Value
    wq: KWorkQueue


class KWorker(Value):
    """struct kworker_s"""

    pid: Value
    work: Value
    wait: Value


class KWorkQueue(Value):
    """struct kwork_wqueue_s"""

    expired: Value
    pending: Value
    sem: Value
    exsem: Value
    nthreads: int
    exit: bool
    worker: List[Value]


class HPWorkQueue(KWorkQueue):
    """struct hp_wqueue_s"""


class LPWorkQueue(KWorkQueue):
    """struct lp_wqueue_s"""



############################################################################
# tools/pynuttx/nxgdb/quickjs.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse

import gdb

from . import autocompeletion, backtrace, utils


@autocompeletion.complete
class QjsDump(gdb.Command):
    """Dump the information of backtrace of JSObject in QuickJS"""

    def parse_arguments(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "-a",
            "--address",
            type=str,
            metavar="symbol",
            default=None,
            help="Variable or address of a JSObject.",
        )
        return parser

    def __init__(self):
        if utils.has_field("JSObject", "backtrace"):
            super().__init__("qjsdump", gdb.COMMAND_USER)
            self.parser = self.parse_arguments()

    def parse_args(self, arg):
        try:
            return self.parser.parse_args(gdb.string_to_argv(arg))
        except SystemExit:
            return

    @utils.dont_repeat_decorator
    def invoke(self, arg: str, from_tty: bool):
        if not (args := self.parse_args(arg)):
            return
        qjs_object = self.get_object(args.address)
        if qjs_object is None:
            gdb.write("invalid jsobject ptr\n")
            return

        if not utils.has_field(qjs_object, "backtrace"):
            gdb.write("this object has not a backtrace\n")
            return

        backtrace = qjs_object["backtrace"]
        jsnative_backtrace = backtrace["js_backtrace"].string()
        gdb.write(f"### dump JSObject backtrace {qjs_object.address} ###\n")
        gdb.write(f"js native backtrace:\n {jsnative_backtrace}\n")
        gdb.write("C/C++ backtrace: \n")
        self.print_jsbacktrace_stack(backtrace)

    def get_object(self, address: str) -> gdb.Value:
        if not address:
            return None

        addr = utils.Value(utils.parse_arg(address))
        return addr.cast(utils.lookup_type("JSObject").pointer()) if addr else None

    def print_jsbacktrace_stack(self, js_backtrace: gdb.Value):
        if utils.has_field(js_backtrace, "backtrace") is not None:
            backtrace_ptr = js_backtrace["backtrace"]
            cnt = int(js_backtrace["cnt"])

            if cnt == 0:
                gdb.write("No backtrace entries found.\n")
                return

            formatter = "{:>1} {:>4} {:>12} {:>12} {:>12} {:>9} {:>14} {:>18} {:}\n"
            leading = formatter.format("", "", "", "", "", "", "", "", "")[:-1]
            btformat = leading + "{1:<48}{2}\n"

            for bt in utils.ArrayIterator(backtrace_ptr, maxlen=cnt):
                gdb.write("BackTrace Entry:\n")
                gdb.write(
                    f"{backtrace.Backtrace(list(utils.ArrayIterator(bt['buffers'], bt['nptrs'])), formatter=btformat)}\n"
                )



############################################################################
# tools/pynuttx/nxstub/utils.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

try:
    from construct import Array, Int8ul, Int16ul, Struct
except ImportError:
    print('Package missing, please do "pip install construct"')

from typing import List

from nxelf.elf import LiefELF


def get_ncpus(elf: LiefELF) -> int:
    # FAR struct tcb_s *g_running_tasks[CONFIG_SMP_NCPUS];
    #
    # g_running_tasks is an pointer array in length of ncpu
    running_tasks = elf.get_symbol("g_running_tasks")

    return running_tasks.size // elf.get_pointer_size()


def get_regsize(elf: LiefELF) -> int:
    """Register size in context"""
    sym = elf.get_symbol("g_last_regs")
    return sym.size // get_ncpus(elf)


def get_tcbinfo(elf: LiefELF):
    tcbinfo_s = Struct(
        "pid_off" / Int16ul,  # FIXME: only little endian supported
        "state_off" / Int16ul,
        "pri_off" / Int16ul,
        "name_off" / Int16ul,
        "stack_off" / Int16ul,
        "stack_size_off" / Int16ul,
        "regs_off" / Int16ul,
        "regs_num" / Int16ul,
    )

    _, data = elf.read_symbol("g_tcbinfo")
    return tcbinfo_s.parse(data)


def get_tcb_size(elf: LiefELF) -> int:
    # static struct tcb_s g_idletcb[CONFIG_SMP_NCPUS];
    # Idle TCB happen to be an array of tcb_s

    idletcb = elf.get_symbol("g_idletcb")
    ncpus = get_ncpus(elf)
    return idletcb.size // ncpus


def parse_array(data, type_, narray):
    return Array(narray, type_).parse(data)


def get_statenames(elf: LiefELF) -> List[str]:
    pointer = elf.get_pointer_type()
    sym, addr = elf.read_symbol("g_statenames")
    names = parse_array(addr, pointer, sym.size // pointer.sizeof())
    names = [elf.read_string(name) for name in names]
    return names


def uint16_t(data: bytes) -> int:
    return Int16ul.parse(data)


def uint8_t(data: bytes) -> int:
    return Int8ul.parse(data)


def get_packet(sock) -> bytes:
    buffer = bytearray()
    started = False
    escaping = False
    checksum = 0
    while True:
        c = sock.recv(1)
        if not started:
            if c in (b"\x03", b"+", b"-"):  # Special packets
                return c
            if c == b"$":
                started = True
            continue

        if escaping:
            c = chr(ord(c) ^ 0x20)
            escaping = False
        elif c == b"}":
            escaping = True
            checksum += ord(c)
            continue

        if c == b"#":
            expected = sock.recv(2)
            expected = int(expected.decode("ascii"), 16)
            if expected != checksum & 0xFF:
                checksum = 0
                started = False
                buffer = bytearray()
                continue
            else:
                break
        else:
            checksum += ord(c)
            buffer.append(ord(c))
    return buffer


def encode_packet(packet: bytes) -> bytes:
    output = list()
    for c in packet:
        if c in b"$#*}":
            output.append(ord("}"))
            c ^= 0x20
        output.append(c)
    return bytes(output)



############################################################################
# tools/pynuttx/nxgdbmcp/src/gmcp/tools/utils.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from typing import Optional

from mcp.server.fastmcp import Context

from ..context import get_session


def register_util_tools(gdb_mcp):
    @gdb_mcp.tool()
    async def gdb_disassemble(
        ctx: Context, session_id: str, location: Optional[str] = None, count: int = 10
    ) -> str:
        """Disassemble code at the specified location or current program counter.

        Args:
            session_id: The GDB session identifier
            location: Optional string specifying where to disassemble
                (e.g., "main", "*0x12345678", "file.c:123").
                If None, disassembles at current PC.
            count: Number of instructions to disassemble (currently not used by the implementation)
        """
        try:
            session = get_session(ctx, session_id)
            command = "disassemble"
            if location:
                command += f" {location}"
            output = await session.execute_command(command)
            return f"Disassembly{f' of {location}' if location else ''}:\n\n{output}"

        except ValueError as e:
            return str(e)
        except Exception as e:
            return f"Failed to disassemble: {str(e)}"

    @gdb_mcp.tool()
    async def gdb_help(
        ctx: Context, session_id: str, command: Optional[str] = None
    ) -> str:
        """Get help for GDB commands.

        Args:
            session_id: The GDB session identifier
            command: Optional string containing the GDB command name to get help for
                (e.g., "break", "print", "step").
                If None, displays general GDB help overview.
        """
        try:
            session = get_session(ctx, session_id)
            if command:
                output = await session.execute_command(f"help {command}")
                return f"Help for '{command}':\n\n{output}"
            else:
                output = await session.execute_command("help")
                return f"GDB help overview:\n\n{output}"

        except ValueError as e:
            return str(e)
        except Exception as e:
            return f"Failed to get help: {str(e)}"



############################################################################
# tools/pynuttx/nxgdb/protocols/fs.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from .value import Value


class Fd(Value):
    """struct fd"""

    f_file: Value
    f_cloexec: Value
    f_tag_fdcheck: Value
    f_tag_fdsan: Value
    f_backtrace: Value


class File(Value):
    """struct file"""

    f_oflags: Value
    f_refs: Value
    f_pos: Value
    f_inode: Value
    f_priv: Value
    f_locked: Value


class Inode(Value):
    """struct inode"""

    i_parent: Value
    i_peer: Value
    i_child: Value
    i_crefs: Value
    i_flags: Value
    u: Value
    i_ino: Value
    i_size: Value
    i_mode: Value
    i_owner: Value
    i_group: Value
    i_atime: Value
    i_mtime: Value
    i_ctime: Value
    i_private: Value
    i_name: Value


class FdList(Value):
    """struct fdlist"""

    fl_rows: Value
    fl_fds: Value



############################################################################
# tools/pynuttx/nxgdbmcp/src/gmcp/tools/shell.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from mcp.server.fastmcp import Context

from ..utils import _exec_command


def register_shell_tools(gdb_mcp):
    @gdb_mcp.tool()
    async def nxgdb_dmesg(ctx: Context, session_id: str) -> str:
        """View device RAM logs, including ramlog and (if available) rpmsg syslog.

        This tool retrieves RAM log information from the device to facilitate debugging and analysis
        of system runtime status. The log content includes ramlog and rpmsg syslog (if supported by the device).

        Args:
            ctx (Context): The context object containing runtime/session information.
            session_id (str): Identifier for the GDB debugging session.

        Returns:
            str: String containing the RAM log content.
        """
        return await _exec_command(ctx, session_id, "dmesg")

    @gdb_mcp.tool()
    async def nxgdb_ps(ctx: Context, session_id: str) -> str:
        """Display a snapshot of current processes and threads.

        Executes the `ps` command to retrieve process and thread information using the
        NuttX thread model, providing a Linux ps-like view of the system state.

        Args:
            ctx (Context): The context object containing runtime/session information.
            session_id (str): Identifier for the GDB debugging session.

        Returns:
            str: Process snapshot showing thread information from the NuttX thread model
                 rather than standard GDB thread listings.
        """
        return await _exec_command(ctx, session_id, "ps")



############################################################################
# tools/pynuttx/nxgdb/fs/fatfs.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import nxgdb.utils as utils
from nxgdb.fs.inode import Inode


def dump_fatfs_cache(node: Inode, path):
    mpt = node.i_private.cast(utils.lookup_type("struct fatfs_mountpt_s").pointer())
    g_drv = utils.parse_and_eval("g_drv").address
    driver = g_drv[mpt.pdrv].cast(utils.lookup_type("struct fatfs_driver_s").pointer())
    opt = driver.drv.u.i_bops.cast(
        utils.lookup_type("struct block_operations").pointer()
    )
    print(f"Fatfs {path} mount point information: {hex(mpt)}")
    print(mpt.dereference().format_string(pretty_structs=True, styling=True))
    print(f"Fatfs {path} driver information: {hex(driver)}")
    print(f"ratio {hex(driver.ratio)}\nblock operations:")
    print(opt.dereference().format_string(pretty_structs=True, styling=True))


def dump_fatfs_file(pointer: utils.Value):
    filep = pointer.cast(utils.lookup_type("struct file").pointer())
    print(f"Fatfs filep information: {hex(filep)}")
    print(filep.dereference().format_string(pretty_structs=True, styling=True))
    fp = filep.f_priv.cast(utils.lookup_type("struct fatfs_file_s").pointer())
    print(f"Fatfs file information: {hex(fp)}")
    print(fp.dereference().format_string(pretty_structs=True, styling=True))



############################################################################
# tools/pynuttx/tests/test_runtime_thread.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import unittest

import gdb
from nxgdb import utils

# The following test cases require running the program as
# we need to access the memory of the program


class TestThread(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pass

    def test_info_threads(self):
        gdb.execute("info nxthreads")

    def test_thread(self):
        gdb.execute("nxthread")

    def test_thread_bad_args(self):
        out = gdb.execute("nxthread apply", to_string=True)
        self.assertEqual(out, "Please specify a thread ID list\n")

        out = gdb.execute("nxthread apply bad", to_string=True)
        self.assertEqual(out, "Please specify a command following the thread ID list\n")

        out = gdb.execute("nxthread apply badid0 cmd", to_string=True)
        self.assertEqual(out, "Please specify a thread ID list and command\n")

        out = gdb.execute("nxthread badid", to_string=True)
        self.assertEqual(out, "Invalid thread id badid\n")

        out = gdb.execute("nxthread 256", to_string=True)
        self.assertEqual(out, "Invalid thread id 256\n")

    def test_thread_apply_all(self):
        out = gdb.execute("nxthread apply all bt", to_string=True)
        self.assertTrue("#0" in out and "Thread" in out, msg=f"Got: {out}")

    def test_thread_apply_with_ids(self):
        out = gdb.execute("nxthread apply 0 bt", to_string=True)
        self.assertTrue("#0" in out and "Thread 0" in out, msg=f"Got: {out}")

        out = gdb.execute("nxthread apply 0 1 info reg", to_string=True)
        # info reg has format like
        # reg      xxx      xxx
        self.assertGreaterEqual(len(out.split("\n")[0].split()), 2, msg=f"Got: {out}")

    def test_thread_with_id(self):
        # This command suppose to switch the stack frames
        # make sure we have at least two threads
        if gdb.parse_and_eval("g_npidhash") < 2:
            out = gdb.execute("nxthread 0", to_string=True)
            self.assertEqual(out, "")
            return

        # Get the current running thread first, after testing we should switch back to it
        # and then continue, otherwise, we might fail on an assertion. For example, we will
        # fail to continue running a task wich is waiting for a mutex if the current running
        # task was actually the idle task.

        cur_thread_id = gdb.parse_and_eval(
            f"g_running_tasks[{gdb.selected_thread().num - 1}]"
        )["pid"]

        gdb.execute("nxthread 0")
        gdb.execute("frame 0")
        cur_sp = utils.get_sp()

        sps = []
        for i in range(gdb.parse_and_eval("g_npidhash")):
            gdb.execute(f"nxthread {i}")
            # switch frame here
            gdb.execute("frame 0")
            new_sp = utils.get_sp()
            sps.append(new_sp)

        # We should have some different stack pointers
        self.assertFalse(all([int(sp) == int(cur_sp) for sp in sps]))

        gdb.execute(f"nxthread {cur_thread_id}")



############################################################################
# tools/pynuttx/nxgdb/critmon.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from dataclasses import dataclass
from typing import Optional

import gdb

from . import utils

CONFIG_SYSTEM_CRITMONITOR = utils.lookup_type("struct critmon_state_s") is not None


@dataclass
class CriticalData:
    """Data class to hold critical data information"""

    pid: Optional[int] = None
    name: Optional[str] = None
    premp_max: Optional[int] = None
    premp_max_caller: Optional[int] = None
    crit_max: Optional[int] = None
    crit_max_caller: Optional[int] = None
    busywait_start: Optional[int] = None
    busywait_caller: Optional[int] = None
    busywait_max: Optional[int] = None
    busywait_max_caller: Optional[int] = None
    busywait_total: Optional[int] = None
    run_max: Optional[int] = None
    run_time: Optional[int] = None

    def __str__(self):
        """Format the information for output"""

        def fmt(field, hex: bool = False):
            val = getattr(self, field)
            if val is None:
                return "-"
            if hex and isinstance(val, int):
                return hex(val)
            return str(val)

        formatter = (
            "{:<3}    {:<18}     "
            "{:<10} {:<9}        "
            "{:<11} {:<9}        "
            "{:<13} {:<6}        "
            "{:<11} {:<9}        "
            "{:<11}        "
            "{:<12}        {}"
        )

        return formatter.format(
            fmt("pid"),
            fmt("name"),
            fmt("premp_max"),
            fmt("premp_max_caller", hex=True),
            fmt("crit_max"),
            fmt("crit_max_caller", hex=True),
            fmt("busywait_start"),
            fmt("busywait_caller", hex=True),
            fmt("busywait_max"),
            fmt("busywait_max_caller", hex=True),
            fmt("busywait_total"),
            fmt("run_max"),
            fmt("run_time"),
        )


class Critmon(gdb.Command):
    """Dump critical resource busy-wait time"""

    def __init__(self):
        if not CONFIG_SYSTEM_CRITMONITOR:
            print("Critmon is not enabled in the current configuration")
            return

        super().__init__("critmon", gdb.COMMAND_USER)

    def process_task_critical_data(self, tcb):
        """Generate information for a single task's critical section"""
        try:
            data = CriticalData()
            for name in utils.get_fieldnames("struct tcb_s"):
                if hasattr(CriticalData, name):
                    value = getattr(tcb, name)
                    if name == "name":
                        value = value.cast(gdb.lookup_type("char").pointer()).string()
                    setattr(data, name, value)
            return data
        except Exception as e:
            print(f"Error processing task critical section information: {str(e)}")
            return None

    def process_global_critical_data(self):
        """Generate information for global critical sections"""
        try:
            output = []
            for cpu, data in enumerate(
                [CriticalData() for _ in range(utils.get_ncpus())]
            ):
                setattr(data, "name", f"CPU{cpu}")
                for key in ("premp_max", "crit_max", "busywait_max", "busywait_total"):
                    value = utils.gdb_eval_or_none(f"g_{key}")
                    if not value:
                        continue
                    setattr(data, key, int(value[cpu]))
                output.append(data)
            return output
        except Exception as e:
            print(f"Error processing global critical section information: {str(e)}")
            return []

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        # header for critical section output
        print(
            "PID    NAME                   "
            "Preemption MaxCaller        "
            "CritSection MaxCaller        "
            "BusywaitStart Caller        "
            "BusywaitMax MaxCaller        "
            "BusywaitAll        "
            "ThreadRunMax        ThreadRunAll"
        )

        try:
            # process global critical data
            for data in self.process_global_critical_data():
                print(data)

            # process task critical data
            for tcb in utils.get_tcbs():
                if not (data := self.process_task_critical_data(tcb)):
                    continue
                print(data)
        except Exception as e:
            print(f"Error invoking critmon command: {str(e)}")

    def diagnose(self, *args, **kwargs):
        return {
            "title": "Critmon Report",
            "summary": "Critical resource busy-wait time dump",
            "result": "info",
            "command": "critmon",
            "category": utils.DiagnoseCategory.sched,
            "data": gdb.execute("critmon", to_string=True),
        }



############################################################################
# tools/pynuttx/nxgdb/wdog.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from __future__ import annotations

from typing import List

import gdb

from . import lists, utils
from .protocols import wdog as p
from .utils import Value


class WDog(Value, p.WDog):
    def __init__(self, wdog: p.WDog):
        if wdog.type.code == gdb.TYPE_CODE_PTR:
            wdog = wdog.dereference()
        super().__init__(wdog)

    def __repr__(self) -> str:
        return (
            f"WDog@{self.address:#x}: tick: {self.expired: <16}"
            f" {self.func.format_string(styling=True)} arg: {self.arg:#x}"
        )

    def __str__(self) -> str:
        return self.__repr__()


def get_wdog_list() -> List[WDog]:
    wdogs = []
    active = utils.parse_and_eval("g_wdactivelist")
    for wdog in lists.NxList(active, "struct wdog_s", "node"):
        wdogs.append(WDog(wdog))

    return wdogs


class WDogDump(gdb.Command):
    """Show wdog timer information"""

    def __init__(self):
        super().__init__("wdog", gdb.COMMAND_USER)

    @utils.dont_repeat_decorator
    def invoke(self, arg, from_tty):
        for wdog in get_wdog_list():
            print(wdog)

    def diagnose(self, *args, **kwargs):
        return {
            "title": "Wdog Timer Information",
            "summary": "Wdog timer information",
            "command": "wdog",
            "result": "info",
            "category": utils.DiagnoseCategory.system,
            "message": gdb.execute("wdog", to_string=True),
        }



############################################################################
# tools/pynuttx/nxgdbmcp/src/gmcp/session/context.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import Dict, Optional

from mcp.server.fastmcp import Context, FastMCP

from .gdb_session import GdbSession


class SessionManager:
    _instance: Optional["SessionManager"] = None

    def __init__(self):
        if SessionManager._instance is not None:
            raise RuntimeError("SessionManager is a singleton")
        self.sessions: Dict[str, GdbSession] = {}

    @classmethod
    def get_instance(cls) -> "SessionManager":
        if cls._instance is None:
            cls._instance = SessionManager()
        return cls._instance

    async def cleanup_all(self):
        for session_id, session in list(self.sessions.items()):
            await session.cleanup()
            self.sessions.pop(session_id, None)


@dataclass
class AppContext:
    """Application context for storing active sessions"""

    @property
    def sessions(self) -> Dict[str, GdbSession]:
        return SessionManager.get_instance().sessions


@asynccontextmanager
async def app_lifespan(server: FastMCP):
    session_manager = SessionManager.get_instance()  # noqa: F841
    try:
        yield AppContext()
    finally:
        pass


def get_session(ctx: Context, session_id: str) -> GdbSession:
    """Get GDB session by ID"""
    sessions = ctx.request_context.lifespan_context.sessions
    if session_id not in sessions:
        raise ValueError(f"No active GDB session with ID: {session_id}")
    return sessions[session_id]



#!/usr/bin/env python3
############################################################################
# tools/pynuttx/etm2human.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse
import logging
import sys

from nxtrace.etmv4_decoder import (
    ATOM_TYPE_E,
    ETBStream,
    ETMv4Decoder,
    ETMv4Packet,
    ETMv4PacketType,
    decode_etb,
)

log = logging.getLogger()


class ETM2Human:

    def emit(self, msg):
        print(msg)
        sys.stdout.flush()

    def trace_info(self, pkt: ETMv4Packet):
        cc_enabled = pkt["cycle_count_enabled"]
        cond_inst_enabled = pkt["conditional_non_branch_instruction_enabled"]
        load_inst_enabled = pkt["load_instruction_enabled"]
        store_inst_enabled = pkt["store_instruction_enabled"]

        self.emit(
            f"TraceInfo - Cycle count {'enabled' if cc_enabled else 'disabled'},\n"
            f"            Tracing of conditional non-branch instruction {'enabled' if cond_inst_enabled else 'disabled'},\n"
            f"            {'NO' if not load_inst_enabled else ''} Explicit tracing of load instructions,\n"
            f"            {'NO' if not store_inst_enabled else ''} Explicit tracing of store instructions,\n"
            f"            p0_key = 0x{pkt['p0_key']:x},\n"
            f"            curr_spec_depth = {pkt['curr_spec_depth']},\n"
            f"            cc_threshold = 0x{pkt['cc_threshold']:x}"
        )

    def trace_on(self, pkt: ETMv4Packet):
        self.emit("TraceOn - A discontinuity in the trace stream")

    def trace_off(self, pkt: ETMv4Packet):
        self.emit("TraceOff - A discontinuity in the trace stream")

    def trace_discard(self, pkt: ETMv4Packet):
        self.emit("Discard")

    def trace_overflow(self, pkt: ETMv4Packet):
        self.emit("Overflow")

    def trace_ts(self, pkt: ETMv4Packet):
        timestamp = hasattr(pkt, "timestamp") and pkt["timestamp"] or 0
        self.emit(f"Timestamp - {timestamp}")
        if hasattr(pkt, "cycle"):
            self.emit(
                f"            (number of cycles between the most recent Cycle Count element {pkt['cycle']})"
            )

    def trace_exception(self, pkt: ETMv4Packet):
        self.emit(
            f"Exception - exception type {pkt['exception_type_name']}, address 0x{pkt['address']:016x}"
        )

    def trace_exception_return(self, pkt: ETMv4Packet):
        self.emit("Exception return")

    def trace_cc(self, pkt: ETMv4Packet):
        if hasattr(pkt, "unknown"):
            self.emit("Cycle count - unknown")
        else:
            self.emit(f"Cycle count - {pkt['cycle']}")

    def trace_commit(self, pkt: ETMv4Packet):
        self.emit(f"Commit - {pkt['commit']}")

    def trace_cancel(self, pkt: ETMv4Packet):
        self.emit(f"Cancel - {pkt['cancel']}")

    def trace_mispredict(self, pkt: ETMv4Packet):
        self.emit("Mispredict")

    def trace_cond_inst(self, pkt: ETMv4Packet):
        log.warning("Conditional instruction is not implemented yet")

    def trace_cond_flush(self, pkt: ETMv4Packet):
        self.emit("Conditional flush")

    def trace_cond_result(self, pkt: ETMv4Packet):
        log.warning("Conditional result is not implemented yet")

    def trace_context(self, pkt: ETMv4Packet):
        setattr(self, "sixty_four_bit", pkt["sixty_four_bit"])
        setattr(self, "ex_level", pkt["ex_level"])
        setattr(self, "security", pkt["security"])

        self.emit(f"Context - Context ID = 0x{pkt['context_id']:x},")
        self.emit(f"          VMID = 0x{pkt['vmid']:x},")
        self.emit(f"          Exception level = EL{pkt['ex_level']},")
        self.emit(f"          Security = {'S' if pkt['security'] else 'NS'},")
        self.emit(f"          {64 if pkt['sixty_four_bit'] else 32}-bit instruction")

    def trace_address(self, pkt: ETMv4Packet):
        if self.sixty_four_bit:
            self.emit(
                f"Address - Instruction address 0x{pkt['address']:016x}, Instruction set Aarch64"
            )
        else:
            if pkt["instruction_set"]:
                self.emit(
                    f"Address - Instruction address 0x{pkt['address']:016x}, Instruction set Aarch32 (Thumb)"
                )
            else:
                self.emit(
                    f"Address - Instruction address 0x{pkt['address']:016x}, Instruction set Aarch32 (ARM)"
                )

    def trace_atom(self, pkt: ETMv4Packet):
        self.emit(f"ATOM - {'E' if pkt['atom_type'] == ATOM_TYPE_E else 'N'}")

    def trace_q(self, pkt: ETMv4Packet):
        if pkt["cycle"]:
            self.emit(f"Q - {pkt['cycle']} of instructions")
        else:
            self.emit("Q - UNKNOWN of instructions")

    etm2human_decoder = {
        ETMv4PacketType.TRACE_INFO: trace_info,
        ETMv4PacketType.TRACE_ON: trace_on,
        ETMv4PacketType.TRACE_OFF: trace_off,
        ETMv4PacketType.DISCARD: trace_discard,
        ETMv4PacketType.OVERFLOW: trace_overflow,
        ETMv4PacketType.TIMESTAMP: trace_ts,
        ETMv4PacketType.EXCEPTION: trace_exception,
        ETMv4PacketType.EXCEPTION_RETURN: trace_exception_return,
        ETMv4PacketType.CYCLE_COUNT: trace_cc,
        ETMv4PacketType.COMMIT: trace_commit,
        ETMv4PacketType.CANCEL: trace_cancel,
        ETMv4PacketType.MISPREDICT: trace_mispredict,
        ETMv4PacketType.COND_INST: trace_cond_inst,
        ETMv4PacketType.COND_FLUSH: trace_cond_flush,
        ETMv4PacketType.COND_RESULT: trace_cond_result,
        ETMv4PacketType.CONTEXT: trace_context,
        ETMv4PacketType.ADDRESS: trace_address,
        ETMv4PacketType.ATOM: trace_atom,
        ETMv4PacketType.Q: trace_q,
    }

    def decode_trace(self, pkt: ETMv4Packet):
        self.etm2human_decoder[pkt.type](self, pkt)


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "-i", "--input", required=True, type=str, help="Give the trace file"
    )
    parser.add_argument(
        "-u",
        "--unaligned",
        action="store_true",
        help="Trace is unaligned and needs to be aligned by frame synchronization packet",
    )
    parser.add_argument(
        "-d",
        "--debug",
        type=str,
        choices=["debug", "info", "warning", "error", "critical"],
        default="warning",
        help="Set the debug level",
    )

    return parser.parse_args()


def main():
    args = parse_args()
    logger = logging.getLogger()
    logger.setLevel(args.debug.upper())

    etm2human = ETM2Human()
    etm_decoder = ETMv4Decoder()
    stream = ETBStream(etm_decoder, args.input, etm2human.decode_trace)
    decode_etb(stream, args.unaligned)


if __name__ == "__main__":
    main()



############################################################################
# tools/pynuttx/nxgdb/protocols/value.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from __future__ import annotations

from typing import Protocol

import gdb


class Value(Protocol):
    address: Value
    is_optimized_out: bool
    type: gdb.Type
    dynamic_type: gdb.Type
    is_lazy: bool
    bytes: bytes

    def cast(self, type: gdb.Type) -> Value: ...
    def dereference(self) -> Value: ...
    def referenced_value(self) -> Value: ...
    def reference_value(self) -> Value: ...
    def rvalue_reference_value(self) -> Value: ...
    def const_value(self) -> Value: ...
    def dynamic_cast(self, type: gdb.Type) -> Value: ...
    def reinterpret_cast(self, type: gdb.Type) -> Value: ...

    def format_string(
        self,
        raw: bool = ...,
        pretty_arrays: bool = ...,
        pretty_structs: bool = ...,
        array_indexes: bool = ...,
        symbols: bool = ...,
        unions: bool = ...,
        address: bool = ...,
        deref_refs: bool = ...,
        actual_objects: bool = ...,
        static_members: bool = ...,
        max_elements: int = ...,
        max_depth: int = ...,
        repeat_threshold: int = ...,
        format: str = ...,
    ) -> str: ...

    def string(
        self, encoding: str = ..., errors: str = ..., length: int = ...
    ) -> str: ...



############################################################################
# tools/pynuttx/nxgdb/gcore.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse
import os
import shutil

import gdb

from . import autocompeletion, mm, utils


def create_file_with_size(filename, size):
    with open(filename, "wb") as f:
        f.write(b"\0" * size)


@autocompeletion.complete
class NXGcore(gdb.Command):

    def get_argparser(self):
        parser = argparse.ArgumentParser()
        parser.add_argument("-o", "--output", metavar="file", help="Gcore output file")
        parser.add_argument(
            "-p",
            "--prefix",
            help="Prefix for objcopy tool",
            default="",
            type=str,
        )

        parser.add_argument(
            "--trust-readonly",
            help="Trust readonly section from elf, bypass read from device.",
            action="store_true",
            default=True,
        )

        parser.add_argument(
            "--no-trust-readonly",
            help="Do not trust readonly section from elf, read from device.",
            action="store_false",
            dest="trust_readonly",
        )

        parser.add_argument(
            "-r",
            "--memrange",
            type=str,
        )
        return parser

    def __init__(self):
        self.tempfile = utils.import_check(
            "tempfile",
            errmsg="No tempfile module found, please try gdb-multiarch instead.\n",
        )

        if not self.tempfile:
            return

        super().__init__("nxgcore", gdb.COMMAND_USER)
        self.parser = self.get_argparser()

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        try:
            args = self.parser.parse_args(gdb.string_to_argv(args))
        except SystemExit:
            return

        objfile = gdb.current_progspace().objfiles()[0]
        elffile = objfile.filename
        tmpfile = self.tempfile.NamedTemporaryFile(suffix=".elf")

        # Create temporary ELF file with sections for each memory region

        shutil.copy(elffile, tmpfile.name)

        # If no parameters are passed in

        if args.output:
            corefile = args.output
        else:
            corefile = elffile + ".core"

        # Get memory range from parameter or elf
        memrange = mm.get_memrange(args.memrange)

        for i, (start, end) in enumerate(memrange):
            # Create a random section name

            section = tmpfile.name + f"{i // 3}"

            # Add objcopy insertion segment command and modify segment start address command

            create_file_with_size(section, end - start)

            gdb.write(f"Using objcopy: {args.prefix}objcopy")
            os.system(
                f"{args.prefix}objcopy --add-section {section}={section} "
                f"--set-section-flags {section}=noload,alloc {tmpfile.name}"
            )
            os.system(
                f"{args.prefix}objcopy --change-section-address "
                f"{section}={hex(start)} {tmpfile.name}"
            )

            os.remove(section)

        gdb.execute(f"file {tmpfile.name}")

        if args.trust_readonly:
            gdb.execute("set trust-readonly-sections on")

        gdb.execute(f"gcore {corefile}")

        if args.trust_readonly:
            # Restore trust-readonly-sections to default off state
            gdb.execute("set trust-readonly-sections off")

        gdb.execute(f'file "{elffile}"')
        tmpfile.close()

        print(f"Please run gdbserver.py to parse {corefile}")



#!/usr/bin/env python3
############################################################################
# tools/pynuttx/traceparse.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse
import logging
import signal
import sys

from nxelf.elf import ELFParser
from nxtrace.rtt import SeggerRTT
from nxtrace.trace import NoteParser


def note_binary_parser(elf_parser, note_parser, binary_file):
    note_parser.parse_file(binary_file)
    note_parser.dump()
    note_parser.flush()


def rtt_parser(elf_parser, note_parser, device, interface, speed, channel):
    # Get the address of the _SEGGER_RTT symbol
    address = elf_parser.symbol_addr("_SEGGER_RTT")
    running = True
    if address is None:
        raise RuntimeError(
            "Symbol _SEGGER_RTT not found, please check if the ELF is correct or if SEGGER RTT is enabled"
        )

    rtt = SeggerRTT(
        device,
        interface,
        speed=speed,
        channel=channel,
        address=address,
    )

    def signal_handler(signum, frame):
        nonlocal running
        if running:
            print("\nReceived signal\n")
            running = False
            rtt.stop()
            note_parser.flush()
            sys.exit(0)
        else:
            print("Already received signal, exiting...")
            sys.exit(0)

    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)

    rtt.start()
    print("Starting RTT...")
    while running:
        data = rtt.read()
        if not data:
            continue

        notes = note_parser.parse(data)
        note_parser.dump(notes)


def arg_parse():
    parser = argparse.ArgumentParser()
    parser.add_argument("-e", "--elf", help="ELF file", required=True)
    parser.add_argument("-o", "--output", help="Output file", default="trace.perfetto")
    parser.add_argument("-b", "--binary", help="Binary note file")
    parser.add_argument("-d", "--device", help="RTT device")
    parser.add_argument("-i", "--interface", help="RTT interface")
    parser.add_argument("-s", "--speed", type=int, help="RTT speed")
    parser.add_argument("-c", "--channel", type=int, help="RTT channel")
    parser.add_argument("-v", "--verbose", action="store_true", help="Verbose output")
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    # Validate the arguments
    if args.binary:
        return vars(args)
    elif args.device and args.interface:
        return vars(args)
    elif args.device or args.interface:
        print("Error: Both --device and --interface must be specified together.")
        sys.exit(1)
    else:
        print(
            "Error: You must specify either a binary file or both RTT parameters (--device and --interface)."
        )
        sys.exit(1)


if __name__ == "__main__":
    args = arg_parse()

    elf_parser = ELFParser(args["elf"])
    note_parser = NoteParser(elf_parser, output=args["output"])

    if args.get("binary"):
        note_binary_parser(elf_parser, note_parser, args["binary"])
    elif args.get("device"):
        rtt_parser(
            elf_parser,
            note_parser,
            args["device"],
            args["interface"],
            args.get("speed"),
            args.get("channel"),
        )



############################################################################
# tools/pynuttx/nxtrace/general_rtt.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import logging
from abc import ABC, abstractmethod
from ctypes import Structure, c_char, c_int32, c_uint32, c_uint64, c_void_p, sizeof
from multiprocessing import shared_memory
from typing import Union

try:
    import lauterbach.trace32.rcl as trace32
    from pylink.jlink import JLink
except ImportError:
    print("Please execute the following command to install dependencies:")
    print("pip install lauterbach-trace32-rcl pylink-square")
    exit(1)

log = logging.getLogger(__name__)


class SEGGER_RTT_CB(Structure):
    _fields_ = [
        ("acID", c_char * 16),
        ("MaxNumUpBuffers", c_int32),
        ("MaxNumDownBuffers", c_int32),
    ]


class SEGGER_RTT_BUFFER_UP(Structure):
    _fields_ = [
        ("sName", c_void_p),
        ("pBuffer", c_void_p),
        ("SizeOfBuffer", c_uint32),
        ("WrOff", c_uint32),
        ("RdOff", c_uint32),
        ("Flags", c_uint32),
    ]


class SEGGER_RTT_BUFFER_DOWN(Structure):
    _fields_ = [
        ("sName", c_void_p),
        ("pBuffer", c_void_p),
        ("SizeOfBuffer", c_uint32),
        ("WrOff", c_uint32),
        ("RdOff", c_uint32),
        ("Flags", c_uint32),
    ]


def struct_factory(cls, bitwidth=32):
    class MeteStructure(type):
        def __new__(cls, name, bases, attrs, bitwidth=32):
            if "_fields_" in attrs:
                realfields = list()
                for field in attrs["_fields_"]:
                    if field[1] == c_void_p:
                        field = (field[0], c_uint32 if bitwidth == 32 else c_uint64)
                    realfields.append(field)
            else:
                raise ValueError("No _fields_ in class")

            class DynamicStructure(Structure):
                _fields_ = realfields
                fields = {field[0]: field[1] for field in realfields}

            return DynamicStructure

    class DynamicStructure(metaclass=MeteStructure, bitwidth=bitwidth):
        _fields_ = cls._fields_

    return DynamicStructure


class MemoryController(ABC):
    @abstractmethod
    def read(self, addr: int, size: int) -> bytes:
        """Read data from memory"""
        pass

    @abstractmethod
    def write(self, addr: int, data: Union[bytes, int]) -> int:
        """Write data to memory"""
        pass


# This class is used to read and write to a structure in memory
class Value:
    def __init__(
        self,
        controller: MemoryController,
        struct: type(Structure),
        addr: int = None,
    ):
        if addr is None:
            raise ValueError("Must specify addr")

        self.struct = struct
        self.controller = controller
        self.address = addr

    def read(self):
        data = self.controller.read(self.address, sizeof(self.struct))
        if len(data) == 0:
            return b""
        parsed = self.struct.from_buffer_copy(data)
        for field, _ in self.struct._fields_:
            self._cache[field] = getattr(parsed, field)

    def __getattr__(self, key):
        if key in self.struct.fields:
            offset = getattr(self.struct, key).offset
            size = getattr(self.struct, key).size
            data = self.controller.read(self.address + offset, size)
            parsed = self.struct.fields[key].from_buffer_copy(data).value
            log.debug("parsed %s: %s", key, parsed)
            return parsed
        else:
            return super().__getattribute__(key)

    def __setattr__(self, key, value):
        if key == "struct":
            super().__setattr__(key, value)
        elif key in self.struct.fields:
            offset = getattr(self.struct, key).offset
            self.controller.write(self.address + offset, value)
        else:
            super().__setattr__(key, value)


class RTTUpChannel:
    def __init__(self, up: Value):
        self.up = up

    def read(self):
        address = self.up.pBuffer
        wroff = self.up.WrOff
        rdoff = self.up.RdOff

        if wroff == rdoff:
            return b""
        elif wroff > rdoff:
            data = self.up.controller.read(address + rdoff, wroff - rdoff)
            self.up.RdOff = wroff
            return data
        else:
            size = self.up.SizeOfBuffer - rdoff
            data = self.up.controller.read(address + rdoff, size)
            data += self.up.controller.read(address, wroff)
            self.up.RdOff = wroff
            return data


class RTTDownChannel:
    def __init__(self, down: Value):
        self.down = down

    def write(self, data: bytes, blocking: bool = False):
        if blocking:
            while data:
                sent = self.write(data, blocking=False)
                data = data[sent:]
            return

        address = self.down.pBuffer
        wroff = self.down.WrOff
        rdoff = self.down.RdOff
        size = self.down.SizeOfBuffer

        if wroff >= size or rdoff >= size:
            raise ValueError("Invalid down buffer")

        written = 0
        if wroff >= rdoff:
            remaining = size - wroff
            if rdoff == 0:
                remaining -= 1

            data_to_write = data[:remaining]
            self.down.controller.write(address + wroff, data_to_write)
            written = len(data_to_write)
            data = data[written:]
            wroff = (wroff + written) % size

        remaining = rdoff - wroff - 1
        if remaining < 0:
            remaining = 0

        num = min(remaining, len(data))
        self.down.controller.write(address + wroff, data[:num])
        written += num
        wroff += num

        self.down.WrOff = wroff
        return written


class SEGGER_RTT:
    def __init__(self, memory: MemoryController, addr: int = None, bitwidth: int = 32):
        self.SEGGER_RTT_CB = struct_factory(SEGGER_RTT_CB, bitwidth)
        self.SEGGER_RTT_BUFFER_UP = struct_factory(SEGGER_RTT_BUFFER_UP, bitwidth)
        self.SEGGER_RTT_BUFFER_DOWN = struct_factory(SEGGER_RTT_BUFFER_DOWN, bitwidth)

        self.rtt_cb = Value(memory, self.SEGGER_RTT_CB, addr=addr)
        self.upbuffer = [bytes() for _ in range(self.rtt_cb.MaxNumUpBuffers)]
        self.downbuffer = [bytes() for _ in range(self.rtt_cb.MaxNumDownBuffers)]

        self.aUp = [
            RTTUpChannel(
                Value(
                    memory,
                    self.SEGGER_RTT_BUFFER_UP,
                    self.rtt_cb.address
                    + sizeof(self.SEGGER_RTT_BUFFER_UP) * i
                    + sizeof(self.SEGGER_RTT_CB),
                )
            )
            for i in range(self.rtt_cb.MaxNumUpBuffers)
        ]

        self.aDown = [
            RTTDownChannel(
                Value(
                    memory,
                    self.SEGGER_RTT_BUFFER_DOWN,
                    self.rtt_cb.address
                    + sizeof(self.SEGGER_RTT_BUFFER_DOWN) * i
                    + sizeof(self.SEGGER_RTT_BUFFER_UP) * self.rtt_cb.MaxNumUpBuffers
                    + sizeof(self.SEGGER_RTT_CB),
                )
            )
            for i in range(self.rtt_cb.MaxNumDownBuffers)
        ]

    def read(self, channel: int, size: int = -1) -> bytes:
        if channel in range(self.rtt_cb.MaxNumUpBuffers):
            self.upbuffer[channel] += self.aUp[channel].read()
            if size == -1:
                data = self.upbuffer[channel]
                self.upbuffer[channel] = bytes()
            else:
                data = self.upbuffer[channel][:size]
                self.upbuffer[channel] = self.upbuffer[channel][size:]
            return data
        else:
            raise ValueError("Invalid channel")

    def write(self, channel: int, data: bytes, blocking: bool = False):
        if channel in range(self.rtt_cb.MaxNumDownBuffers):
            self.aDown[channel].write(data, blocking)
        else:
            raise ValueError("Invalid channel")


class Trace32MemoryController(MemoryController):
    def __init__(self, debugger: trace32.Debugger):
        self.debugger = debugger
        self._addr = dict()

    def read(self, addr: int, size: int) -> bytes:
        if addr not in self._addr:
            self._addr[addr] = self.debugger.address(access="E", value=addr)

        ret = self.debugger.memory.read_uint8_array(
            self._addr[addr], length=size
        ).tobytes()
        log.debug("trace32 read %#x len:%d %s", addr, size, ret)
        return ret

    def write(self, addr: int, data: Union[bytes, int]) -> int:
        if addr not in self._addr:
            self._addr[addr] = self.debugger.address(access="D", value=addr)

        if isinstance(data, int):
            value = data.to_bytes(length=4, byteorder="little")
        else:
            value = data

        ret = self.debugger.memory.write_uint8_array(self._addr[addr], value)
        log.debug("trace32 write %#x len:%d %s", addr, len(value), ret)
        return ret


class JlinkMemoryController(MemoryController):
    def __init__(self, debugger: JLink):
        self.debugger = debugger

    def read(self, addr: int, size: int) -> bytes:
        ret = bytes(self.debugger.memory_read8(addr=addr, num_bytes=size))
        log.debug("jlink read %#x len:%d %s", addr, size, ret)
        return ret

    def write(self, addr: int, data: Union[bytes, int]) -> int:
        if isinstance(data, int):
            data = data.to_bytes(length=4, byteorder="little")

        ret = self.debugger.memory_write8(addr=addr, data=data)
        log.debug("jlink write %#x len:%d %s", addr, len(data), ret)
        return ret


class ShmMemoryController(MemoryController):
    def __init__(self, name: str, rttaddr: int):
        self.shm = shared_memory.SharedMemory(name=name)
        self.rttaddr = rttaddr
        self.rttoffset = self.shm.buf.tobytes().find(b"SEGGER RTT")

    def read(self, addr: int, size: int) -> bytes:
        offset = addr - self.rttaddr + self.rttoffset
        if offset < 0 or offset + size > self.shm.size:
            raise ValueError("Invalid address")

        ret = self.shm.buf[offset : offset + size].tobytes()
        log.debug(f"shm read 0x{addr:x}(offset:{offset}) size:{size} {ret}")
        return ret

    def write(self, addr: int, data: Union[bytes, int]) -> int:
        offset = addr - self.rttaddr + self.rttoffset
        if isinstance(data, int):
            data = data.to_bytes(length=4, byteorder="little")

        if offset < 0 or offset + len(data) > self.shm.size:
            raise ValueError("Invalid address")

        self.shm.buf[offset : offset + len(data)] = data
        log.debug(f"shm write 0x{addr:x}(offset:{offset}) size:{len(data)} {data}")
        return len(data)



############################################################################
# tools/pynuttx/nxgdbmcp/src/gmcp/tools/command.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from mcp.server.fastmcp import Context

from ..utils import _exec_command


def register_command_tools(gdb_mcp):

    @gdb_mcp.tool()
    async def gdb_command(ctx: Context, session_id: str, command: str) -> str:
        """Execute a GDB command"""
        return await _exec_command(ctx, session_id, command)

    @gdb_mcp.tool()
    async def gdb_get_nxgdb_cmd_list(ctx: Context, session_id: str) -> str:
        """Get nxgdb command help info, nxgdb is a set of gdb scripts for debugging NuttX.
        Note that nxgdb may have issues. When analyzing NuttX crashes,
        you need to determine whether the problem is with nxgdb or with NuttX itself.
        """
        return await _exec_command(ctx, session_id, "help user-defined")



############################################################################
# tools/pynuttx/tests/test_runtime_profile.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import unittest

import gdb

# The following test cases require running the program as
# we need to access the memory of the program


class TestProfile(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pass

    def check_output(self, out, expect=""):
        if expect not in out:
            self.fail(f"Got: {out}")

    def test_profile(self):

        out = gdb.execute(" profile p 0x123", to_string=True)
        self.check_output(out, expect="function calls in")
        self.check_output(out, expect="_gdb.execute")
        # Example of output:

    # (gdb) profile p 0x123
    # $3 = 291
    #          12 function calls in 0.001 seconds

    #    Ordered by: cumulative time

    #    ncalls  tottime  percall  cumtime  percall filename:lineno(function)
    #         1    0.000    0.000    0.000    0.000 {built-in method builtins.exec}
    #         1    0.000    0.000    0.000    0.000 <string>:1(<module>)
    #         1    0.000    0.000    0.000    0.000 {built-in method _gdb.execute}
    #         1    0.000    0.000    0.000    0.000 {method 'disable' of '_lsprof.Profiler' objects}

    def test_time(self):
        # (gdb) time p 0x123
        # $4 = 291
        # Time elapsed: 0.000295s

        out = gdb.execute("time", to_string=True)
        self.check_output(out, expect="Time elapsed")



############################################################################
# tools/pynuttx/nxgdb/tlsdump.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import argparse

import gdb

from . import autocompeletion, utils
from .lists import NxDQueue

CONFIG_TLS_NELEM = utils.get_field_nitems("struct tls_info_s", "tl_elem")


@autocompeletion.complete
class TlsDump(gdb.Command):
    """Dump and check the integrity of tls_info and task_info"""

    def get_argparser(self):
        parser = argparse.ArgumentParser(description=self.__doc__)
        parser.add_argument(
            "-p",
            "--pid",
            type=int,
            help="Dump the tls info of the thread/task represented by this pid",
            default=None,
        )
        parser.add_argument(
            "-c",
            "--check",
            action="store_true",
            help="Integrity check.All threads in a task must share the same task_info",
        )
        return parser

    def __init__(self):
        if not CONFIG_TLS_NELEM:
            print("TLS is not enabled in the current configuration")
            return
        super().__init__("tlsdump", gdb.COMMAND_USER)
        self.parser = self.get_argparser()

    def parse_arguments(self, argv):
        try:
            args = self.parser.parse_args(argv)
        except SystemExit:
            return None

        return args

    def check_corruption(self, tcb):
        """integrity check"""

        try:
            tcb_s = utils.lookup_type("struct tcb_s").pointer()
            if not tcb or tcb.type != tcb_s:
                print("No tcb found, or the tcb type is invalid")
                return True

            # Get the task_info of the task, and compare it with the task_info of each thread
            # If the tcb is a task, get the task_info of the task
            # If the tcb is a thread, get the task_info of the task to which the thread belongs
            task = (
                tcb
                if utils.get_tcb_type(tcb) == "TASK"
                else utils.get_tcb(tcb.group.tg_pid)
            )
            if not task or not utils.get_tcb_type(task):
                print("Can not find the task within the group")
                return True

            tls_info_s = utils.lookup_type("struct tls_info_s").pointer()
            task_info = task.stack_alloc_ptr.cast(tls_info_s).tl_task
            corrupted = False
            # Traverse all threads under this task through the group linked list
            for tcb in NxDQueue(task.group.tg_members, "struct tcb_s", "member"):
                # Get the task_info of this thread
                info = tcb.stack_alloc_ptr.cast(tls_info_s).tl_task
                # Only report when corrupted
                if info != task_info:
                    pid = int(tcb.pid)
                    print(
                        f"PID:{pid} is corrupted, task_info addr:{hex(task_info)}, got {hex(info)}"
                    )
                    corrupted = True
            return corrupted
        except gdb.error:
            print("Error occurred during integrity check")
            return True

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        args = self.parse_arguments(gdb.string_to_argv(args))
        if not args:
            return

        # tlsdump -c
        # Do integrity check
        if args.check:
            corrupted = any(self.check_corruption(tcb) for tcb in utils.get_tcbs())
            print(f"Check: {'FAILED' if corrupted else 'PASS'}")
            return

        # tlsdump / tlsdump -p pid
        tls_info_s = utils.lookup_type("struct tls_info_s").pointer()
        CONFIG_TLS_TASK_NELEM = utils.get_field_nitems("struct task_info_s", "ta_telem")
        pid = args.pid
        tcbs = [utils.get_tcb(pid)] if pid is not None else utils.get_tcbs()
        for tcb in tcbs:
            if not tcb or not (task_type := utils.get_tcb_type(tcb)):
                continue
            tls_info = tcb.stack_alloc_ptr.cast(tls_info_s)
            task_info = tls_info.tl_task
            print(f"PID:{tcb.pid}, {task_type}, task_info addr:{hex(task_info)}")

            if task_type == "TASK":
                print("task tls elements:")
                for i in range(CONFIG_TLS_TASK_NELEM):
                    tls = task_info.ta_telem[i]
                    print(f"{i}: {hex(tls)}")

            print("thread tls elements:")
            for i in range(CONFIG_TLS_NELEM):
                tls = tls_info.tl_elem[i]
                print(f"{i}: {hex(tls)}")

    def diagnose(self, *args, **kwargs):
        corrupted = any(self.check_corruption(tcb) for tcb in utils.get_tcbs())

        return {
            "title": "tlsdump report",
            "summary": "integrity check",
            "result": "failed" if corrupted else "pass",
            "category": utils.DiagnoseCategory.system,
            "command": "tlsdump",
            "data": gdb.execute("tlsdump", to_string=True),
        }



############################################################################
# tools/pynuttx/nxgdb/protocols/uorb.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from __future__ import annotations

from .value import Value


class OrbMetadata(Value):
    """struct orb_metadata_s"""

    o_name: Value
    o_size: Value
    o_format: Value


class SensorMeta(Value):
    """struct sensor_meta_s"""

    esize: Value
    name: Value


class SensorState(Value):
    """struct sensor_state_s"""

    esize: Value
    nbuffer: Value
    min_latency: Value
    min_interval: Value
    nsubscribers: Value
    nadvertisers: Value
    generation: Value
    priv: Value


class SensorUState(Value):
    """struct sensor_ustate_s"""

    esize: Value
    latency: Value
    interval: Value
    generation: Value


class SensorUpper(Value):
    """struct sensor_upperhalf_s"""

    lower: Value
    state: SensorState
    timing: Value
    buffer: Value
    lock: Value
    userlist: Value


class SensorLower(Value):
    """struct sensor_lowerhalf_s"""

    type: Value
    nbuffer: Value
    uncalibrated: Value
    ops: Value
    push_event: Value
    notify_event: Value

    sensor_lock: Value
    sensor_unlock: Value
    priv: Value
    persist: Value


class SensorUser(Value):
    """struct sensor_user_s"""

    node: Value
    fds: Value
    role: Value
    changed: Value
    event: Value
    flushing: Value
    buffersem: Value
    bufferpos: Value
    state: SensorUState



############################################################################
# tools/pynuttx/nxgdbmcp/src/gmcp/session/gdb_session.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from typing import Optional

import gdbrpc


class GdbSession:
    def __init__(
        self,
        session_id: str,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):

        self.id = session_id
        self.host = host if host else "localhost"
        self.port = port if port else 20819
        self.session = gdbrpc.Client(self.host, self.port)

        assert self.session.connect()

    async def execute_command(self, command: str) -> str:
        if not self.session:
            raise RuntimeError("GDB socket connection not established")

        try:
            response = self.session.call(gdbrpc.ShellExec(command))
            return response

        except Exception as e:
            raise RuntimeError(f"Error executing command '{command}': {str(e)}")

    async def cleanup(self):
        try:
            if self.session:
                self.session.disconnect()
        except Exception as e:
            print(f"Error during cleanup: {e}")
        finally:
            self.session = None
            print("GDB session cleanup completed")



############################################################################
# tools/pynuttx/nxgdb/autocompeletion.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from __future__ import annotations

import gdb

metavar_map = {"file": gdb.COMPLETE_FILENAME, "symbol": gdb.COMPLETE_SYMBOL}


# The parser._actions type is `class _HelpAction`
# you can refer https://github.com/python/cpython/blob/3.10/Lib/argparse.py#L1091
def get_options(parser, filter=None, member="option_strings"):
    rs = []
    for action in parser._actions:
        if filter is None or filter(action):
            # option_strings: A list of command-line option strings which
            #   should be associated with this action. ['-a', '--arch']
            # dest: The name of the attribute to hold the created object(s). 'arch'
            value = getattr(action, member)
            if value and isinstance(value, list):
                rs.extend(value)
            elif value:
                rs.append(value)
    return rs


def get_option_choices(parser, name):
    for action in parser._actions:
        if action.dest == name:
            return action.choices


def get_option_metavar(parser, name):
    for action in parser._actions:
        if action.dest == name:
            return action.metavar


def get_options_metavar_filter(name):
    return lambda action: action.metavar == name


def get_options_choices_filter():
    return lambda action: action.choices is not None


def get_options_positional_filter():
    return lambda action: not action.option_strings


def complete_for_options(parser, word):
    argv = gdb.string_to_argv(word)
    if word == "" or word.endswith(" "):
        argv.append("")
    # gcc -march=native, the `march` is option, `native` is argument
    latest_option = argv[-2] if len(argv) > 1 else argv[0]
    latest_argument = argv[-1]

    options = get_options(parser)
    # `--help` -> long_option, `help` -> long_option_without_dash
    long_options_without_dash = [
        item[2:] for item in options if len(item) > 2 and item[:2] == "--"
    ]

    positional_options = get_options(
        parser, get_options_positional_filter(), member="dest"
    )
    pos_opt = (
        metavar_map.get(get_option_metavar(parser, positional_options[0]))
        if len(positional_options) != 0
        else None
    )

    return (
        options,
        long_options_without_dash,
        pos_opt,
        latest_option,
        latest_argument,
    )


def complete_specific_argument(parser, option):
    for meta_var, completion_list in metavar_map.items():
        options = get_options(parser, get_options_metavar_filter(meta_var))
        if option in options:
            return completion_list
    return None


def common_argument_completion(
    parser,
    word,
):
    """
    Common command line argument completion for custom gdb commands.
    """

    options, long_options_without_dash, pos_opt, latest_option, latest_argument = (
        complete_for_options(parser, word)
    )

    choice_options = get_options(parser, get_options_choices_filter(), member="dest")
    choices = [get_option_choices(parser, choice) for choice in choice_options]
    option_without_dash = ""
    if latest_option[:2] == "--":
        option_without_dash = latest_option[2:]
    elif latest_option[:1] == "-":
        option_without_dash = latest_option[1:]
    for index, element in enumerate(choice_options):
        if option_without_dash == element:
            return [item for item in choices[index] if item.startswith(latest_argument)]

    # to complete arguments
    rs = complete_specific_argument(parser, latest_option)
    if rs:
        return rs

    # to complete options
    if latest_argument == "-":
        return [item[1:] for item in options]
    if latest_argument.startswith("--"):
        return [
            item
            for item in long_options_without_dash
            if item.startswith(latest_argument[2:])
        ]
    if latest_argument.startswith("-"):
        return [latest_argument[1:]]

    # If the option is completed, but no arguments is provided.
    if latest_option and latest_option[0] == "-":
        return []
    # First, we can try to complete positional options
    if pos_opt:
        return pos_opt
    return options


def complete(self):
    def complete(self, word, text):
        return common_argument_completion(self.parser, word)

    self.complete = complete
    return self



############################################################################
# tools/pynuttx/tests/test_runtime_fs.py
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import unittest

import gdb

# The following test cases require running the program as
# we need to access the memory of the program


class TestFs(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        pass

    def check_output(self, out, expect=""):
        if expect not in out:
            self.fail(f"Got: {out}")

    def test_fdinfo(self):
        out = gdb.execute("fdinfo", to_string=True)
        self.check_output(out, expect="PID: 1")

    def test_fdinfo_pid(self):
        out = gdb.execute("fdinfo -p 1", to_string=True)
        self.check_output(out, expect="PID: 1")

    def test_mount(self):
        out = gdb.execute("mount", to_string=True)
        self.check_output(out, expect="/proc type procfs")

    def test_foreach_inode(self):
        out = gdb.execute("foreach inode", to_string=True)
        self.check_output(out, expect=" console ")

    def test_foreach_inode_addr(self):
        pass

    def test_info_shm(self):
        pass



############################################################################
# tools/pynuttx/nxgdb/nxcrash/deadlock.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from __future__ import annotations

import gdb

from .. import utils


class DeadLock(gdb.Command):
    """Detect and report if threads have deadlock."""

    def __init__(self):
        super().__init__("deadlock", gdb.COMMAND_USER)

    def has_deadlock(self, pid):
        """Check if the thread has a deadlock"""
        tcb = utils.get_tcb(pid)
        if not tcb or not tcb["waitobj"]:
            return False

        sem = tcb["waitobj"].cast(utils.lookup_type("sem_t").pointer())
        if not utils.sem_is_mutex(sem):
            return False

        # It's waiting on a mutex
        holder = utils.mutex_get_holder(sem)
        if holder in self.holders:
            return True

        self.holders.append(holder)
        return self.has_deadlock(holder)

    def collect(self, tcbs):
        """Collect the deadlock information"""

        detected = []
        collected = []
        for tcb in tcbs:
            self.holders = []  # Holders for this tcb
            pid = tcb["pid"]
            if pid in detected or not self.has_deadlock(tcb["pid"]):
                continue

            # Deadlock detected
            detected.append(pid)
            detected.extend(self.holders)
            collected.append((pid, self.holders))

        return collected

    def diagnose(self, *args, **kwargs):
        collected = self.collect(utils.get_tcbs())

        return {
            "title": "Deadlock Report",
            "summary": f"{'No' if not collected else len(collected)} deadlocks",
            "result": "fail" if collected else "pass",
            "category": utils.DiagnoseCategory.sched,
            "command": "crash deadlock",
            "deadlocks": {int(pid): [i for i in h] for pid, h in collected},
        }

    @utils.dont_repeat_decorator
    def invoke(self, args, from_tty):
        collected = self.collect(utils.get_tcbs())
        if not collected:
            gdb.write("No deadlock detected.\n")
            return

        for pid, holders in collected:
            gdb.write(f'Thread {pid} "{utils.get_task_name(pid)}" has deadlocked!\n')
            gdb.write(f"  holders: {pid}->")
            gdb.write("->".join(str(pid) for pid in holders))
            gdb.write("\n")



############################################################################
# tools/pynuttx/nxstub/target.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import logging
import traceback
from typing import List

from nxreg.register import Registers

from . import utils

# Manually limit the maximum number of threads, in case the memory is
# corrupted to non-sense values.
MAX_THREADS = 512


class RawMemory:
    def __init__(self, address: int, data: bytearray):
        self.address = address
        self.data = bytearray(data)

    def __len__(self):
        return len(self.data)

    def __str__(self):
        return f"Memory({self.address:#x}~{self.address + len(self.data):#x})"

    def __repr__(self):
        return self.__str__()


class ThreadInfo:
    def __init__(self, name, pid, state, registers: Registers):
        self.name = name
        self.pid = pid
        self.state = state
        self.registers = registers

    def __str__(self):
        return f"{self.name}({self.pid}) {self.state}"

    def __repr__(self):
        return f"Thread({self.name}, {self.pid}, {self.state})"


class Target:
    PID0_REPLACE = 0x7FFFFFFF

    def __init__(
        self,
        elf,
        arch=None,
        registers: Registers = None,
        memories: List[RawMemory] = None,
        remap=None,
        core=None,
    ):
        """
        The target that GDB stub operations on.
        :param elf: The ELF file path.
        :param arch: The architecture of the target, e.g. "arm", "riscv", "mips", etc.
        :param registers: The optional initial register value, normally used for crash log analysis.
        :param memories: The optional initial memory regions, normally used for raw memory dump.
        :param remap: The optional memory remap table, used to remap memory regions. A list of tuple (from, to, length).
        :param core: The optional core dump file path.
        """
        self.logger = logging.getLogger(__name__)
        self.elf: utils.LiefELF = elf
        self.core: utils.LiefELF = core
        self.registers = registers or Registers(elf, arch=arch)
        self.memories: List[RawMemory] = []
        self.registers.readmem = self.memory_read
        self.arch = arch
        self.pid = self.PID0_REPLACE  # Current thread PID
        self.remap = remap or []
        for mem in memories or []:
            # Go through the write process to merge overlapping memory regions
            self.memory_write(mem.address, mem.data)

        self.logger.debug(f"Memory regions: {self.memories}")

    def _read_symbol(self, symbol: str, length: int = 0) -> bytes:
        sym = self.elf.get_symbol(symbol)
        data = self.memory_read(sym.value, length or sym.size)
        return data, sym

    def _read_int(self, symbol: str) -> int:
        inttype = self.elf.get_inttype()
        data, sym = self._read_symbol(symbol, inttype.sizeof())
        if not data:
            return None, sym
        return inttype.parse(data), sym

    def _read_str(self, address: int) -> str:
        output = b""
        while (b := self.memory_read(address, 1)) != b"\0":
            output += b
            address += 1

        return output.decode("utf-8", errors="replace")

    def update_threads(self) -> List[ThreadInfo]:
        """Update the latest threads information"""

        self.threads = (
            ThreadInfo("main", self.PID0_REPLACE, "Running", self.registers),
        )

        try:
            pointer = self.elf.get_pointer_type()
            tcbsize = utils.get_tcb_size(self.elf)
            tcbinfo = utils.get_tcbinfo(self.elf)
            try:
                states = utils.get_statenames(self.elf)
            except Exception:
                # Fallback to pure number. Don't bother to parse value of NUM_TASK_STATES, 256 is enough.
                states = [str(i) for i in range(256)]

            npidhash, sym = self._read_int("g_npidhash")
            self.logger.debug(f"g_npidhash: {npidhash}@{sym.value:#x}")
            if not npidhash:
                self.logger.error(f"No threads info found: {npidhash}")
                return self.threads

            if npidhash > MAX_THREADS:
                self.logger.warning(f"g_npidhash looks wrong: {npidhash}")
                npidhash = MAX_THREADS

            ncpus = utils.get_ncpus(self.elf)
            regsize = utils.get_regsize(self.elf)

            data, sym = self._read_symbol("g_running_tasks")  # an array of pointers
            running_tasks = utils.parse_array(data, pointer, ncpus)
            self.logger.debug(f"g_running_tasks: {running_tasks}@{sym.value:#x}")

            def parse_tcb(address: int) -> ThreadInfo:
                registers = Registers(
                    self.elf, arch=self.arch, readmem=self.memory_read
                )
                data = self.memory_read(address, tcbsize)
                if not data or len(data) != tcbsize:
                    self.logger.error(f"Invalid TCB size: {len(data)} != {tcbsize}")
                    return ThreadInfo("<invalid>", 0, "Invalid", registers)

                if tcbinfo.name_off == 0:
                    name = "<noname>"
                else:
                    name = self._read_str(address + tcbinfo.name_off)
                self.logger.debug(f"loading thread: {name}, tcb@{address:#x}")
                pid = utils.uint16_t(data[tcbinfo.pid_off : tcbinfo.pid_off + 2])
                pid = pid if pid != 0 else self.PID0_REPLACE
                state = utils.uint8_t(data[tcbinfo.state_off : tcbinfo.state_off + 1])
                state = states[state] if state < len(states) else "Unknown"

                xcpregs = None
                if address in running_tasks:
                    # Running task registers is not in memory, best chance is the registers
                    # stored in g_last_regs when assert happened.
                    last_regs = self.elf.get_symbol("g_last_regs").value
                    cpu = running_tasks.index(address)
                    xcpregs = cpu * regsize + last_regs
                    # Check if the g_last_regs is obviously invalid
                    if all(b == 0 for b in self.memory_read(xcpregs, regsize)):
                        self.logger.error(
                            f"Invalid g_last_regs: @ {xcpregs:#x}, use TCB instead"
                        )
                        xcpregs = None
                    else:
                        print(
                            f"\x1b[31;1mLoad CPU{cpu} task [{name}] registers from g_last_regs\x1b[m"
                        )

                if not xcpregs:
                    off = tcbinfo.regs_off
                    xcpregs = data[off : off + pointer.sizeof()]
                    xcpregs = pointer.parse(xcpregs)

                try:
                    registers.load(addr=xcpregs)
                except ValueError as e:
                    self.logger.error(f"Failed to load registers: {e}")

                self.logger.debug(f"Parse TCB: {name}({pid},{state})")
                return ThreadInfo(name, pid, state, registers)

            data, sym = self._read_symbol("g_pidhash")
            pidhash = pointer.parse(data)
            data = self.memory_read(pidhash, pointer.sizeof() * npidhash)
            pidhash = utils.parse_array(data, pointer, npidhash)
            self.logger.debug(f"g_pidhash: {pidhash}@{sym.value:#x}")

            self.threads = [parse_tcb(tcb) for tcb in pidhash if tcb]
            self.logger.debug(f"Found {self.threads}")
        except Exception as e:
            self.logger.error(f"No threads info: {e}\n{traceback.format_exc()}")

        return self.threads

    def switch_thread(self, pid: int = None) -> Registers:
        """
        Switch to the thread with PID, or the next running thread.
        Return the registers for the thread
        """
        self.logger.debug(f"Switch to thread {pid}")
        if pid == -1 or pid == 0:
            # -1: all threads, 0: current thread. No need to switch for both cases
            # See https://sourceware.org/gdb/current/onlinedocs/gdb.html/Packets.html#Packets
            return self.registers

        if pid is None:
            # None is only used during our initial setup, not of GDB RSP protocol
            pid = self.PID0_REPLACE

        for t in self.threads:
            if pid == t.pid:
                self.pid = pid
                self.registers = t.registers
                return self.registers

    def current_thread(self) -> int:
        return self.pid

    def memory_read(self, address: int, length: int) -> bytes:
        self.logger.debug(f"Read: {address:#x} {length}Bytes")

        # Check the real address from remap table
        for fromaddr, toaddr, total in self.remap:
            if toaddr <= address < toaddr + total:
                address = fromaddr + (address - toaddr)
                if address + length > fromaddr + total:
                    # Do not support cross region read
                    length = fromaddr + total - address
                self.logger.debug(f"Remap to: {address:#x} {length}Bytes")
                break

        # Try cached memory
        for mem in self.memories:
            if mem.address <= address < mem.address + len(mem):
                offset = address - mem.address
                # Limit the length to available data
                length = min(length, len(mem) - offset)
                return mem.data[offset : offset + length]

        # Try core
        if self.core and (value := self.core.read_from(address, length)):
            return bytes(value)

        # Try elf
        return bytes(self.elf.read_from(address, length) or [])

    def memory_write(self, address, data):
        memories = self.memories
        mem = RawMemory(address, bytearray(data))
        self.logger.debug(f"Write: {mem}")

        if not memories or address > memories[-1].address + len(memories[-1]):
            memories.append(mem)  # New memory region in the end
            return
        elif address + len(data) < memories[0].address:
            memories.insert(0, mem)  # New memory region in the beginning
            return

        for i, m in enumerate(memories):
            if address > m.address + len(m):
                continue

            if address + len(data) < m.address:
                memories.insert(i, mem)  # New memory region in the middle
            elif (offset := address - m.address) >= 0:
                # Overwrite and append data to existing memory
                m.data[offset : offset + len(data)] = data
            else:
                # Prepend data to existing memory
                offset = address + len(data) - m.address
                m.data = data + m.data[offset:]
                m.address = address

            # Remove overlapping memory regions
            end = m.address + len(m)
            i = memories.index(m)  # Recalculate index, in case of list changed
            for m2 in memories[i + 1 :]:
                if end < m2.address:
                    break

                self.logger.debug(f"Remove overlapping memory: {m2}")
                memories.remove(m2)
                m.data += m2.data[end - m2.address :]

            break

    def monitor_command(self, command: bytes) -> str:
        """Handle monitor command"""
        self.logger.debug(f"Monitor command: {command}")
        if command.startswith(b"setregs"):
            command = command.decode("ascii")
            _, address = command.split(" ")

            # Check frequently used symbols
            if address.startswith("g_last_regs"):
                # address could be g_last_regs or g_last_regs[1] or g_last_regs[0] etc.
                if "[" in address:
                    regsize = utils.get_regsize(self.elf)
                    splitted = address.split("[")
                    index = int(splitted[1].split("]")[0])
                    address = self.elf.get_symbol(splitted[0]).value
                    address += index * regsize
                else:
                    symbol = self.elf.get_symbol(address)
                    address = symbol.value
            else:
                try:
                    address = (
                        int(address, 16)
                        if "0x" in address or "0X" in address
                        else int(address)
                    )
                except ValueError:
                    # try if it's a symbol, note that expression is not supported.
                    address = self.elf.get_symbol(address).value

            self.registers.load(address)
            return f"Loaded registers from {address:#x}\n"
        elif command.startswith(b"help"):
            return (
                "setregs <address|symbol>: Load registers from address or symbol.\n"
                "\tNeed to execute `maint flush register-cache` to see latest registers\n"
                "help: Show this help message\n"
            )



############################################################################
# tools/pynuttx/nxgdb/fs/littlefs.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import nxgdb.utils as utils
from nxgdb.fs.inode import Inode


def dump_little_cache(node: Inode, path):
    mpt = node.i_private.cast(utils.lookup_type("struct littlefs_mountpt_s").pointer())
    drv = mpt.drv.dereference()
    geo = mpt.geo
    cfg = mpt.cfg
    lfs = mpt.lfs
    print("littlefs {path} mount point information: {hex(mpt)}")
    print("drv:")
    print(drv.format_string(pretty_structs=True, styling=True))
    print("cfg:")
    print(cfg.format_string(pretty_structs=True, styling=True))
    print("geo:")
    print(geo.format_string(pretty_structs=True, styling=True))
    print("lfs:")
    print(lfs.format_string(pretty_structs=True, styling=True))



############################################################################
# tools/pynuttx/nxgdb/protocols/thread.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from .fs import FdList
from .value import Value


class TgBininfo(Value):
    """struct tg_bininfo_s"""

    entrypt: Value
    mapped: Value
    mod: Value
    picbase: Value
    addrenv: Value
    oldenv: Value
    mapsize: Value
    priority: Value
    stacksize: Value
    uid: Value
    gid: Value
    mode: Value
    stackaddr: Value
    unload: Value


class Group(Value):
    """struct group_s"""

    tg_pid: Value
    tg_ppid: Value
    tg_flags: Value
    tg_uid: Value
    tg_gid: Value
    tg_euid: Value
    tg_egid: Value
    tg_members: Value
    tg_bininfo: TgBininfo
    tg_children: Value
    tg_nchildren: Value
    tg_exitcode: Value
    tg_nwaiters: Value
    tg_waitflags: Value
    tg_exitsem: Value
    tg_statloc: Value
    tg_joinlock: Value
    tg_joinqueue: Value
    tg_info: Value
    tg_sigactionq: Value
    tg_sigpendingq: Value
    tg_sigdefault: Value
    tg_envp: Value
    tg_envc: Value
    itimer: Value
    tg_fdlist: FdList
    tg_mm_map: Value


class Tcb(Value):
    """struct tcb_s"""

    flink: Value
    blink: Value
    group: Group
    member: Value
    join_queue: Value
    join_entry: Value
    join_sem: Value
    join_val: Value
    addrenv_curr: Value
    pid: Value
    sched_priority: Value
    init_priority: Value
    start: Value
    entry: Value
    task_state: Value
    boost_priority: Value
    base_priority: Value
    holdsem: Value
    cpu: Value
    affinity: Value
    flags: Value
    lockcount: Value
    errcode: Value
    timeslice: Value
    sporadic: Value
    waitdog: Value
    adj_stack_size: Value
    stack_alloc_ptr: Value
    stack_base_ptr: Value
    dspace: Value
    waitobj: Value
    sigprocmask: Value
    sigwaitmask: Value
    sigpendactionq: Value
    sigpostedq: Value
    sigunbinfo: Value
    ticks: Value
    run_start: Value
    run_max: Value
    run_time: Value
    premp_start: Value
    premp_max: Value
    premp_caller: Value
    premp_max_caller: Value
    crit_start: Value
    crit_max: Value
    crit_caller: Value
    crit_max_caller: Value
    perf_event_ctx: Value
    perf_event_mutex: Value
    xcp: Value
    sigdeliver: Value
    name: Value
    stackrecord_pc: Value
    stackrecord_sp: Value
    stackrecord_pc_deepest: Value
    stackrecord_sp_deepest: Value
    sp_deepest: Value
    caller_deepest: Value
    level_deepest: Value
    level: Value



############################################################################
# tools/pynuttx/gdbrpc/client.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

import logging
import os
import queue
import socket
import sys
import threading
from datetime import datetime
from typing import Dict, Optional, Tuple

import cloudpickle as pickle
from gdbrpc.utils import (
    DEFAULT_TIMEOUT,
    PacketStatus,
    PostRequest,
    Request,
    Response,
    socket_recv,
    socket_send,
)


class Client:
    def __init__(
        self,
        host: str = "localhost",
        port: int = 20819,
        log_level: int = logging.INFO,
        log_path: Optional[str] = None,
    ):
        self._host = host
        self._port = port
        self._socket: socket.socket
        self._connected = False
        self._response = queue.Queue()
        self._pending_requests: Dict[int, PostRequest] = {}

        self._logger = logging.getLogger(__name__)
        if not self._logger.hasHandlers():
            self._logger.setLevel(log_level)

            if log_path is None:
                timestamp = datetime.now().strftime("%Y%m%d-%H%M%S")
                pid = os.getpid()
                log_path = f"gdbrpc_client-{timestamp}-pid{pid}.log"

            handler = logging.FileHandler(log_path)
            formatter = logging.Formatter("%(asctime)s gdbrpc_client: %(message)s")
            handler.setFormatter(formatter)
            self._logger.addHandler(handler)

    def connect(self):
        try:
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self._socket.connect((self._host, self._port))
            self._connected = True
            logging.info(f"Connected to GDB server at {self._host}:{self._port}")

            threading.Thread(
                target=self._listen_responses, daemon=True, args=(self._socket,)
            ).start()

            return True
        except Exception as e:
            logging.error(f"Failed to connect {self._host}:{self._port}: {e}")
            return False

    def _listen_responses(self, socket: socket.socket):
        try:
            while self._connected:
                data: Tuple[Response, PacketStatus] = pickle.loads(
                    socket_recv(socket, self._logger)
                )
                response, status = data
                self._logger.debug(
                    f"Received response for request ID {response.tag}, current queue: {self._pending_requests}"
                )

                if status == PacketStatus.PYTHON_VERSION_MISMATCH:
                    response.payload += f"\nclient python version: {sys.version}"
                    self._response.put(response.payload)
                elif (
                    status == PacketStatus.HAS_CALLBACK
                    and response.tag in self._pending_requests
                ):
                    self._logger.debug(f"Handling callback for request ID {status}")
                    callback = self._pending_requests.get(response.tag)

                    assert isinstance(callback, PostRequest)

                    try:
                        callback(response.payload)
                    except Exception as e:
                        self._logger.error(
                            f"Error in callback {callback}, which raised an exception: {e}"
                        )
                    finally:
                        callback.finish.set()
                else:
                    self._response.put(response)
                    self._logger.debug(
                        f"Response for request ID {response.tag} put into response queue"
                    )
                self._logger.info(
                    f"Received data: {response.payload}. from {self._host}:{self._port}"
                )
        except ConnectionError:
            self._logger.info("Connection closed by server")
            self.disconnect()
        except Exception as e:
            self._logger.error(f"Error receiving data: {e}")
            self.disconnect()

    def disconnect(self):
        self._connected = False
        self._logger.info("Disconnecting...")
        if self._socket:
            try:
                self._socket.close()
                self._logger.info("Socket closed successfully.")
            except Exception as e:
                self._logger.error(f"Error closing socket: {e}")
        self._logger.info("Disconnected")

    def no_pending_requests(self) -> bool:
        return len(self._pending_requests) == 0

    def call(
        self,
        request: Request,
        post_request: Optional[PostRequest] = None,
        timeout: float = DEFAULT_TIMEOUT,
    ):
        if not self._connected:
            raise ConnectionError("Not connected to server")
        if not isinstance(request, Request):
            raise TypeError("request must be a Request instance")
        if post_request is not None and not isinstance(post_request, PostRequest):
            raise TypeError("post_request must be a PostRequest instance or None")

        if post_request is not None:
            self._logger.debug(f"Registering callback for request ID {request.tag}")
            self._pending_requests[request.tag] = post_request
            payload = (request, PacketStatus.HAS_CALLBACK)
        else:
            self._logger.debug(f"No callback for request ID {request.tag}")
            payload = (request, PacketStatus.NO_CALLBACK)

        self._logger.debug(f"Sending request: {request} to {self._host}:{self._port}")

        socket_send(self._socket, pickle.dumps(payload), self._logger)

        try:
            rs = self._response.get(timeout=timeout)
        except queue.Empty:
            self._logger.error("Request timed out")
            raise TimeoutError("Request timed out")

        return rs.payload



############################################################################
# tools/pynuttx/nxgdb/wqueue.py
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.  The
# ASF licenses this file to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance with the
# License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
############################################################################

from __future__ import annotations

from typing import List

import gdb

from . import lists, utils
from .protocols import wqueue as p
from .utils import Value

CONFIG_SCHED_WORKQUEUE = utils.lookup_type("struct kworker_s") is not None


class Work(Value, p.Work):
    def __init__(self, work: p.Work):
        if work.type.code == gdb.TYPE_CODE_PTR:
            work = work.dereference()
        super().__init__(work)

    @property
    def wqueue(self) -> WorkQueue:
        return WorkQueue(self.wq)

    def __repr__(self) -> str:
        return f"(struct work_s *){self.address:#x}: {self.worker.format_string(styling=True)} arg={self.arg}"

    def __str__(self) -> str:
        return self.__repr__()


class KWorker(Value, p.KWorker):
    """Worker thread information"""

    def __init__(self, worker: p.KWorker, wqueue=None):
        if worker.type.code == gdb.TYPE_CODE_PTR:
            worker = worker.dereference()
        super().__init__(worker)
        self.wqueue = wqueue

    @property
    def is_running(self):
        return self.work

    @property
    def work(self) -> Work:
        work = self["work"]
        return Work(work) if work else None

    @property
    def name(self):
        return utils.get_task_name(utils.get_tcb(self.pid))

    def __repr__(self):
        return f"(struct kworker_s *){self.address:#x} {self.work or 'idle'}"

    def __str__(self):
        return self.__repr__()


class WorkQueue(Value, p.KWorkQueue):
    """Work queue information"""

    def __init__(self, queue: p.KWorkQueue, name: str = None):
        if queue.type.code == gdb.TYPE_CODE_PTR:
            queue = queue.dereference()
        super().__init__(queue)
        self.name = name or "<noname>"

    @property
    def expired(self) -> List[Work]:
        work_s = utils.lookup_type("struct work_s")
        return [
            Work(worker.cast(work_s.pointer()))
            for worker in lists.NxList(self["expired"])
        ]

    @property
    def pending(self) -> List[Work]:
        work_s = utils.lookup_type("struct work_s")
        return [
            Work(worker.cast(work_s.pointer()))
            for worker in lists.NxList(self["pending"])
        ]

    @property
    def running(self) -> List[Work]:
        return [thread.work for thread in self.threads if thread.is_running]

    @property
    def workers(self) -> List[Work]:
        return self.expired + self.pending + self.running

    @property
    def threads(self) -> List[KWorker]:
        return [
            KWorker(thread, wqueue=self)
            for thread in utils.ArrayIterator(self.worker, self.nthreads)
        ]

    @property
    def nthreads(self):
        return int(self["nthreads"])

    @property
    def worker(self):
        return utils.Value(
            (utils.sizeof("struct kwork_wqueue_s") + int(self.address))
        ).cast(utils.lookup_type("struct kworker_s").pointer())

    @property
    def is_running(self):
        return any(thread.is_running for thread in self.threads)

    @property
    def is_exiting(self):
        return self.exit

    def __repr__(self):
        state = "running" if self.is_running else "idle"
        return (
            f"{self.name} (struct kwork_wqueue_s *){self.address:#x}, {state}, "
            f"{self.nthreads} threads, {len(self.workers)} work"
        )

    def __str__(self):
        return self.__repr__()


def get_work_queues() -> List[WorkQueue]:
    entry = utils.parse_and_eval("work_thread")
    kwork_wqueue_s = utils.lookup_type("struct kwork_wqueue_s")

    tcbs = utils.get_tcbs()
    #  The function address may be or'ed with 0x01
    tcbs = filter(lambda tcb: int(tcb.entry.main) & ~0x01 == entry, tcbs)
    queue = []
    for tcb in tcbs:
        if not (args := utils.get_task_argvstr(tcb)):
            continue
        #   wqueue  = (FAR struct kwork_wqueue_s *)
        #             ((uintptr_t)strtoul(argv[1], NULL, 16));
        #   kworker = (FAR struct kworker_s *)
        #             ((uintptr_t)strtoul(argv[2], NULL, 16));
        wqueue = gdb.Value(int(args[1], 16)).cast(kwork_wqueue_s.pointer())
        wqueue = WorkQueue(wqueue, name=utils.get_task_name(tcb))
        if wqueue not in queue:
            queue.append(wqueue)

    return queue


class WorkQueueDump(gdb.Command):
    """Show work queue information"""

    def __init__(self):
        if not CONFIG_SCHED_WORKQUEUE:
            return

        super().__init__("worker", gdb.COMMAND_USER)

    @utils.dont_repeat_decorator
    def invoke(self, arg, from_tty):
        queues = get_work_queues()
        for queue in queues:
            print(f"{queue}")
            if queue.is_running:
                print("    Running:")  # Dump the work that is running
                running = [thread for thread in queue.threads if thread.is_running]
                print("\n".join(f"    {thread}" for thread in running))
            else:
                print("    No running work")

            if not queue.workers:
                continue

            expired = queue.expired
            pending = queue.pending
            if expired:
                print("    Expired:")
                print("\n".join(f"    {work}" for work in expired))
            if pending:
                print("    Pending:")
                print("\n".join(f"    {work}" for work in pending))
            print("")

    def diagnose(self, *args, **kwargs):
        return {
            "title": "Worker Queue Information",
            "summary": "All worker queue information",
            "command": "worker",
            "result": "info",
            "category": utils.DiagnoseCategory.system,
            "message": gdb.execute("worker", to_string=True),
        }
